# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from networking_arista._i18n import _

# Resource actions
CREATE = 'create'
DELETE = 'delete'
FULL_SYNC = 'full_sync'

# Resource types
TENANT_RESOURCE = 'tenant'
NETWORK_RESOURCE = 'network'
SEGMENT_RESOURCE = 'segment'
DHCP_RESOURCE = 'dhcp'
ROUTER_RESOURCE = 'router'
VM_RESOURCE = 'vm'
BAREMETAL_RESOURCE = 'baremetal'
PORT_SUFFIX = '_port'
DHCP_PORT_RESOURCE = DHCP_RESOURCE + PORT_SUFFIX
ROUTER_PORT_RESOURCE = ROUTER_RESOURCE + PORT_SUFFIX
VM_PORT_RESOURCE = VM_RESOURCE + PORT_SUFFIX
BAREMETAL_PORT_RESOURCE = BAREMETAL_RESOURCE + PORT_SUFFIX
PORT_BINDING_RESOURCE = 'port_binding'

ALL_RESOURCE_TYPES = [TENANT_RESOURCE,
                      NETWORK_RESOURCE,
                      SEGMENT_RESOURCE,
                      DHCP_RESOURCE,
                      ROUTER_RESOURCE,
                      VM_RESOURCE,
                      BAREMETAL_RESOURCE,
                      DHCP_PORT_RESOURCE,
                      VM_PORT_RESOURCE,
                      BAREMETAL_PORT_RESOURCE,
                      PORT_BINDING_RESOURCE]

# Constants
INTERNAL_TENANT_ID = 'INTERNAL-TENANT-ID'
MECHANISM_DRV_NAME = 'arista'

# SG Constants

# When a SG is applied to a VM, ingress refers to traffic flowing
# into a VM and egress refers to traffic flowing out.
# In the baremetal case, traffic flowing out of a switchport is
# flowing into the baremetal. Therefore, INGRESS SG rules
# should be applied as 'out' ACLs and EGRESS rules as 'in' ACLs.
INGRESS_DIRECTION = 'out'
EGRESS_DIRECTION = 'in'

# EAPI error messages of interest
EOS_UNREACHABLE_MSG = _('Unable to reach EOS')
ERR_CVX_NOT_LEADER = _('only available on cluster leader')



# Copyright (c) 2018 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from oslo_log import fixture as log_fixture

import cProfile
from eventlet import queue
import mock
from pstats import Stats

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as n_const
import neutron_lib.context
from neutron_lib.plugins import constants as p_const
from neutron_lib.plugins import directory
from neutron_lib.services.trunk import constants as trunk_const
from oslo_config import cfg

from neutron.api.rpc.handlers import l3_rpc
from neutron.common import utils as common_utils
from neutron.plugins.ml2.drivers import type_vxlan  # noqa
from neutron.tests.common import helpers
from neutron.tests.unit.plugins.ml2 import test_plugin

from networking_arista.common import db_lib
from networking_arista.ml2 import arista_sync
from networking_arista.tests.unit import utils


ENABLE_PROFILER = False
ENABLE_LOGGING = False


class MechTestBase(test_plugin.Ml2PluginV2TestCase):
    """Main test cases for Arista Mechanism driver.

    Tests all mechanism driver APIs supported by Arista Driver. It invokes
    all the APIs as they would be invoked in real world scenarios and
    verifies the functionality.
    """

    _mechanism_drivers = ['arista', 'openvswitch']

    def get_additional_service_plugins(self):
        p = super(MechTestBase, self).get_additional_service_plugins()
        p.update({'trunk_plugin_name': 'trunk'})
        return p

    def setUp(self):
        if ENABLE_LOGGING:
            log_fixture.setlevel.logging.root.setLevel(logging.DEBUG)
            self.useFixture(log_fixture.SetLogLevel(['neutron'],
                                                    logging.DEBUG))
        if ENABLE_PROFILER:
            self.pr = cProfile.Profile()
            self.pr.enable()
        utils.setup_arista_wrapper_config(cfg)
        cfg.CONF.set_override('vni_ranges',
                              ['10000:11000'],
                              group='ml2_type_vxlan')
        super(MechTestBase, self).setUp()
        self.plugin.notifier.port_update = self._mock_port_update
        self.plugin.start_rpc_listeners()
        self.host1 = 'host1'
        self.host2 = 'host2'
        self.host3 = 'host3'
        # Hack to ensure agents report being alive
        cfg.CONF.set_override('agent_down_time', 1000)
        helpers.register_ovs_agent(
            host=self.host1, bridge_mappings={self.physnet: 'br-eth1'})
        helpers.register_ovs_agent(
            host=self.host2, bridge_mappings={self.physnet: 'br-eth1'})
        helpers.register_ovs_agent(
            host=self.host3, bridge_mappings={self.physnet2: 'br-eth1'})
        self.region = 'region'
        self.cvx = utils.MockCvx(self.region)
        self.drv = self.driver.mechanism_manager.mech_drivers['arista'].obj
        # multiprocessing.Queue's get() fails to wake up a thread, swap
        # it out for a LightQueue for testing purposes
        self.drv.provision_queue = queue.LightQueue()
        for worker in self.driver._workers:
            if isinstance(worker, arista_sync.AristaSyncWorker):
                self.ar_sync = worker
                self.ar_sync._rpc = self.cvx
                self.ar_sync.provision_queue = self.drv.provision_queue
            worker.start()
        self.trunk_plugin = directory.get_plugin('trunk')
        self.net_count = 0

    def tearDown(self):
        for worker in self.driver._workers:
            worker.stop()
            worker.wait()
        self.cvx.endpoint_data.clear()
        super(MechTestBase, self).tearDown()
        if ENABLE_PROFILER:
            p = Stats(self.pr)
            p.strip_dirs()
            p.sort_stats('cumtime')
            p.print_stats()

    def _mock_port_update(self, context, port, network_type, segmentation_id,
                          physical_network):
        '''Simulates an L2 agent's response to a port_update notification

        After binding a port, the ML2 plugin notifies all L2 agents of the
        binding so that they can configure the datapath. Once they have
        done so, the call either update_devices_up or update_devices_list
        which brings the port status to ACTIVE. This function simply calls
        update_devices_list in response to a binding without actually doing
        any datapath manipulation.
        '''
        plugin = directory.get_plugin()
        rpc_plugin = plugin.endpoints[0]
        host = port.get(portbindings.HOST_ID)
        agent_id = 'ovs-agent-%s' % host
        vif_type = port.get(portbindings.VIF_TYPE)
        if port.get('device_owner') == trunk_const.TRUNK_SUBPORT_OWNER:
            return
        if vif_type == 'ovs':
            device_list = []
            device_list.append(port['id'])
            if port.get('trunk_details'):
                trunk_rpc = self.trunk_plugin._rpc_backend._skeleton
                for s in port['trunk_details']['sub_ports']:
                    s['trunk_id'] = port['trunk_details']['trunk_id']
                    trunk_rpc.update_subport_bindings(
                        self.context, port['trunk_details']['sub_ports'])
                    device_list.append(s['port_id'])
            devices_dict = {'devices_up': device_list,
                            'agent_id': agent_id,
                            'host': port.get(portbindings.HOST_ID)}

            # This is a hack. When calling update_port_status from the rpc
            # handler, the trunk_details db extension gets a new admin context
            # in order to query the parent port's subports' mac address.
            # Querying within the new context's session seems to somehow
            # invalidate the transaction in update_port_status, which causes
            # the status in the db to remain 'DOWN' in spite of an
            # update_port_[pre/post]commit being sent indicating that the
            # status is 'ACTIVE'. For now, I've found that using the same admin
            # context in all queries resolves the issue. In my testing, this
            # doesn't affect real environments using mysql and seems to be
            # limited to sqlite
            #
            # Additional notes: If there is no transaction in progress when a
            # query in the new context is issued, the update_port_status
            # commit succeeds (ie. comment out the context.session.flush() in
            # update_individual_port_db_status
            with mock.patch.object(neutron_lib.context, 'get_admin_context',
                                   return_value=self.context):
                rpc_plugin.update_device_list(context, **devices_dict)

            if port.get('trunk_details'):
                trunk_rpc = self.trunk_plugin._rpc_backend._skeleton
                trunk_rpc.update_trunk_status(
                    self.context, port['trunk_details']['trunk_id'], 'ACTIVE')
        elif (port.get(portbindings.VNIC_TYPE) == 'normal'
              and vif_type == 'unbound'):
            device_dict = {'agent_id': agent_id,
                           'device': port['id'],
                           'host': port.get(portbindings.HOST_ID)}
            rpc_plugin.update_device_down(context, **device_dict)

    def create_network(self, net_dict):
        net = self.plugin.create_network(self.context, net_dict)
        self.plugin.create_subnet(self.context,
                                  {'subnet':
                                   {'tenant_id': net['tenant_id'],
                                    'name': net['name'],
                                    'network_id': net['id'],
                                    'ip_version': 4,
                                    'cidr': '10.0.%d.0/24' % self.net_count,
                                    'gateway_ip': '10.0.%d.1' % self.net_count,
                                    'allocation_pools': None,
                                    'enable_dhcp': False,
                                    'dns_nameservers': None,
                                    'host_routes': None}})
        self.net_count += 1
        n_ctxs = self.plugin.get_network_contexts(self.context, [net['id']])
        return net, n_ctxs[net['id']]

    def delete_network(self, net_id):
        self.plugin.delete_network(self.context, net_id)

    def create_port(self, port_dict):
        minimal_port = {'port':
                        {'name': port_dict.get('name'),
                         'tenant_id': port_dict.get('tenant_id'),
                         'device_id': port_dict.get('device_id'),
                         'fixed_ips': port_dict.get('fixed_ips'),
                         'network_id': port_dict.get('network_id'),
                         'device_owner': '',
                         'admin_state_up': True}}
        port = self.plugin.create_port(self.context, minimal_port)
        full_port = {'port': port_dict}
        port = self.plugin.update_port(self.context, port['id'], full_port)
        p_ctx = self.plugin.get_bound_port_context(self.context, port['id'])
        return port, p_ctx

    def migrate_port(self, port_id, new_host):
        port_dict = {'port': {'binding:host_id': new_host}}
        port = self.plugin.update_port(self.context, port_id, port_dict)
        p_ctx = self.plugin.get_bound_port_context(self.context, port_id)
        return port, p_ctx

    def migrate_dhcp_device(self, port_id, new_device):
        port_dict = {'port':
                     {'device_id': n_const.DEVICE_ID_RESERVED_DHCP_PORT}}
        port = self.plugin.update_port(self.context, port_id, port_dict)
        port_dict = {'port': {'device_id': new_device}}
        port = self.plugin.update_port(self.context, port_id, port_dict)
        p_ctx = self.plugin.get_bound_port_context(self.context, port_id)
        return port, p_ctx

    def bind_trunk_to_host(self, port, device_id, host):
        p_dict = {'port':
                  {'device_owner': n_const.DEVICE_OWNER_COMPUTE_PREFIX,
                   portbindings.HOST_ID: host,
                   'device_id': device_id}}
        port = self.plugin.update_port(self.context, port['id'], p_dict)
        p_ctx = self.plugin.get_bound_port_context(self.context, port['id'])
        return port, p_ctx

    def bind_subport_to_trunk(self, port, trunk):
        parent = self.plugin.get_port(self.context, trunk['port_id'])
        p_dict = {'port':
                  {portbindings.HOST_ID: parent.get(portbindings.HOST_ID),
                   'device_owner': trunk_const.TRUNK_SUBPORT_OWNER}}
        port = self.plugin.update_port(self.context, port['id'], p_dict)
        self.plugin.update_port_status(self.context, port['id'],
                                       n_const.PORT_STATUS_ACTIVE)

    def unbind_port_from_host(self, port_id):
        p_dict = {'port':
                  {portbindings.HOST_ID: None,
                   'device_id': ''}}
        port = self.plugin.update_port(self.context, port_id, p_dict)
        return port

    def bind_trunk_to_baremetal(self, port_id, device_id, host,
                                switch_id, switch_port):
        p_dict = {'port':
                  {'device_id': device_id,
                   'device_owner': n_const.DEVICE_OWNER_BAREMETAL_PREFIX,
                   'binding:host_id': host,
                   'binding:profile': {'local_link_information': [
                       {'switch_id': switch_id,
                        'port_id': switch_port}]},
                   'binding:vnic_type': 'baremetal'}}
        port = self.plugin.update_port(self.context, port_id, p_dict)
        return port

    def unbind_trunk_from_baremetal(self, port_id):
        p_dict = {'port':
                  {'device_id': '',
                   'device_owner': '',
                   'binding:host_id': None,
                   'binding:profile': None,
                   'binding:vnic_type': None,
                   'status': n_const.PORT_STATUS_DOWN,
                   portbindings.VIF_TYPE: portbindings.VIF_TYPE_UNBOUND}}
        self.plugin.update_port(self.context, port_id, p_dict)

    def bind_dvr_to_host(self, port, host):
        p_dict = {'port':
                  {'device_id': port['device_id'],
                   'device_owner': port['device_owner'],
                   portbindings.HOST_ID: host}}
        self.plugin.update_distributed_port_binding(self.context,
                                                    port['id'], p_dict)
        self.plugin.update_port_status(self.context, port['id'],
                                       n_const.PORT_STATUS_ACTIVE,
                                       host)
        p_ctx = self.plugin.get_bound_port_context(self.context, port['id'],
                                                   host)
        return port, p_ctx

    def unbind_dvr_from_host(self, port, host):
        self.plugin.update_port_status(
            self.context, port['id'], n_const.PORT_STATUS_DOWN, host)

    def set_port_to_error_state(self, port):
        self.plugin.update_port_status(
            self.context, port['id'], n_const.PORT_STATUS_ERROR)

    def delete_port(self, port_id):
        self.plugin.delete_port(self.context, port_id)

    def _get_endpoint(self, resource_type):
        endpoint_map = {
            'tenant': 'region/%s/tenant' % self.region,
            'network': 'region/%s/network' % self.region,
            'segment': 'region/%s/segment' % self.region,
            'dhcp': 'region/%s/dhcp' % self.region,
            'router': 'region/%s/router' % self.region,
            'vm': 'region/%s/vm' % self.region,
            'baremetal': 'region/%s/baremetal' % self.region,
            'dhcp_port': 'region/%s/port?type=dhcp' % self.region,
            'router_port': 'region/%s/port?type=router' % self.region,
            'vm_port': 'region/%s/port?type=vm' % self.region,
            'baremetal_port': 'region/%s/port?type=baremetal' % self.region,
            'port_binding': 'region/%s/portbinding' % self.region}
        return endpoint_map[resource_type]

    def _assertResourceCreated(self, resource_type, resource_id):
        endpoint = self._get_endpoint(resource_type)

        def resource_created():
            return resource_id in self.cvx.endpoint_data[endpoint].keys()
        common_utils.wait_until_true(resource_created)

    def _assertResourceDeleted(self, resource_type, resource_id):
        endpoint = self._get_endpoint(resource_type)

        def resource_deleted():
            return resource_id not in self.cvx.endpoint_data[endpoint].keys()
        common_utils.wait_until_true(resource_deleted)

    def assertTenantCreated(self, tenant_id):
        self._assertResourceCreated('tenant', tenant_id)

    def assertTenantDeleted(self, tenant_id):
        self._assertResourceDeleted('tenant', tenant_id)

    def assertNetworkCreated(self, network_id):
        self._assertResourceCreated('network', network_id)

    def assertNetworkDeleted(self, network_id):
        self._assertResourceDeleted('network', network_id)

    def assertSegmentCreated(self, segment_id):
        self._assertResourceCreated('segment', segment_id)

    def assertSegmentDeleted(self, segment_id):
        self._assertResourceDeleted('segment', segment_id)

    def assertDhcpCreated(self, instance_id):
        self._assertResourceCreated('dhcp', instance_id)

    def assertDhcpDeleted(self, instance_id):
        self._assertResourceDeleted('dhcp', instance_id)

    def assertRouterCreated(self, instance_id):
        self._assertResourceCreated('router', instance_id)

    def assertRouterDeleted(self, instance_id):
        self._assertResourceDeleted('router', instance_id)

    def assertVmCreated(self, instance_id):
        self._assertResourceCreated('vm', instance_id)

    def assertVmDeleted(self, instance_id):
        self._assertResourceDeleted('vm', instance_id)

    def assertBaremetalCreated(self, instance_id):
        self._assertResourceCreated('baremetal', instance_id)

    def assertBaremetalDeleted(self, instance_id):
        self._assertResourceDeleted('baremetal', instance_id)

    def assertDhcpPortCreated(self, port_id):
        self._assertResourceCreated('dhcp_port', port_id)

    def assertDhcpPortDeleted(self, port_id):
        self._assertResourceDeleted('dhcp_port', port_id)

    def assertRouterPortCreated(self, port_id):
        self._assertResourceCreated('router_port', port_id)

    def assertRouterPortDeleted(self, port_id):
        self._assertResourceDeleted('router_port', port_id)

    def assertVmPortCreated(self, port_id):
        self._assertResourceCreated('vm_port', port_id)

    def assertVmPortDeleted(self, port_id):
        self._assertResourceDeleted('vm_port', port_id)

    def assertBaremetalPortCreated(self, port_id):
        self._assertResourceCreated('baremetal_port', port_id)

    def assertBaremetalPortDeleted(self, port_id):
        self._assertResourceDeleted('baremetal_port', port_id)

    def assertPortBindingCreated(self, pb_key):
        self._assertResourceCreated('port_binding', pb_key)

    def assertPortBindingDeleted(self, pb_key):
        self._assertResourceDeleted('port_binding', pb_key)


class L3HARouterTestFramework(MechTestBase):

    L3Plugin = ('neutron.services.l3_router.l3_router_plugin.'
                'L3RouterPlugin')

    def get_additional_service_plugins(self):
        p = super(L3HARouterTestFramework,
                  self).get_additional_service_plugins()
        p.update({'flavors_plugin_name': 'neutron.services.flavors.'
                                         'flavors_plugin.FlavorsPlugin'})
        p.update({'l3_plugin_name': self.L3Plugin})
        return p

    def _register_l3_agent(self, host):
        return helpers.register_l3_agent(host=host)

    def setUp(self):
        super(L3HARouterTestFramework, self).setUp()
        self.l3_plugin = directory.get_plugin(p_const.L3)
        self.l3_rpc_cb = l3_rpc.L3RpcCallback()

    def create_router(self, ha=False):
        router = self.l3_plugin.create_router(
            self.context, {'router': {'name': 'router',
                                      'admin_state_up': True,
                                      'tenant_id': self._tenant_id,
                                      'ha': ha,
                                      'distributed': False}})
        self.l3_plugin.schedule_router(self.context, router['id'])
        for host in self.l3_agents:
            self.sync_routers(router['id'], host['host'])
        return router

    def sync_routers(self, router_id, host):
        """Call to L3 Agent plugin sync_routers

        Since only l3 agent plugin is enabled in the unit test, to get the
        sync data for routers (which causes port update), we need to call the
        l3 agent plugin sync_routers method directly.
        This call is normally done by a process running periodically on
        l3 agent server (neutron-l3-agent).
        """
        self.l3_rpc_cb.sync_routers(self.context, host=host,
                                    router_ids=[router_id])

    def add_router_interface(self, router, interface_info):
        return self.l3_plugin.add_router_interface(self.context, router['id'],
                                                   interface_info)

    def remove_router_interface(self, router, interface_info):
        return self.l3_plugin.remove_router_interface(self.context,
                                                      router['id'],
                                                      interface_info)

    def delete_router(self, router_id):
        return self.l3_plugin.delete_router(self.context, router_id)

    def get_ha_network(self, router):
        networks = [getattr(net, 'Network') for net in db_lib.get_networks()]
        HA_network_name = n_const.HA_NETWORK_NAME % router['project_id']
        net_id = [net.id for net in networks if net.name == HA_network_name]
        return net_id[0]

    def get_network_segments(self, network_id):
        segments = db_lib.get_segments()
        net_segs = [seg.id for seg in segments if seg.network_id == network_id]
        return net_segs

    def update_routers_states(self, router_id, l3_agent):
        binding_l3_agent_host = l3_agent['host']
        binding_l3_agent_id = l3_agent['id']
        self.l3_plugin.update_routers_states(self.context,
                                             {router_id: 'active'},
                                             binding_l3_agent_host)
        bindings = self.l3_plugin.get_ha_router_port_bindings(self.context,
                                                              [router_id])
        binding = [binding for binding in bindings
                   if binding.l3_agent_id == binding_l3_agent_id][0]
        return self.plugin.get_port(self.context, binding.port_id)

    def get_legacy_router_port(self, port_id):
        return self.plugin.get_port(self.context, port_id)

    def assertL3HANetworkCreated(self, router, net_id):
        endpoint = self._get_endpoint('network')
        HA_network_name = n_const.HA_NETWORK_NAME % router['project_id']

        def resource_created():
            cvx_data = self.cvx.endpoint_data[endpoint]
            return (cvx_data and net_id in cvx_data and
                    HA_network_name == cvx_data[net_id]['name'])
        common_utils.wait_until_true(resource_created)

    def assertLegacyRouterCreated(self, router, host):
        endpoint = self._get_endpoint('router')

        def resource_created():
            cvx_data = list(self.cvx.endpoint_data[endpoint].values())
            expected_data = {'hostId': host,
                             'tenantId': router['project_id'],
                             'id': router['id']}
            return cvx_data and cvx_data[0] == expected_data
        common_utils.wait_until_true(resource_created)

    def assertL3HARouterCreated(self, router):
        endpoint = self._get_endpoint('router')

        def resource_created():
            cvx_data = list(self.cvx.endpoint_data[endpoint].values())
            expected_data = {'hostId': 'HA Router',
                             'tenantId': router['project_id'],
                             'id': router['id']}
            return cvx_data and cvx_data[0] == expected_data
        common_utils.wait_until_true(resource_created)

    def assertL3HAPortCreated(self, router, port_id):
        endpoint = self._get_endpoint('router_port')
        port_name = n_const.HA_PORT_NAME % router['project_id']

        def resource_created():
            cvx_data = self.cvx.endpoint_data[endpoint]
            return (cvx_data and port_id in cvx_data and all(
                [cvx_data[port_id]['tenantId'] == router['project_id'],
                 cvx_data[port_id]['portName'] == port_name,
                 cvx_data[port_id]['instanceId'] == router['id']]))
        common_utils.wait_until_true(resource_created)

    def assertSegmentsCreated(self, segments):
        segment_endpoint = self._get_endpoint('segment')

        def resource_created():
            cvx_segments = self.cvx.endpoint_data[segment_endpoint].keys()
            return cvx_segments and set(cvx_segments) == set(segments)
        common_utils.wait_until_true(resource_created)

    def assertSegmentsDeleted(self, segments):
        segment_endpoint = self._get_endpoint('segment')

        def resource_deleted():
            cvx_data = self.cvx.endpoint_data[segment_endpoint]
            cvx_segments = cvx_data.keys()
            return all(seg not in cvx_segments for seg in segments)
        common_utils.wait_until_true(resource_deleted)

    def assertRouterPortsDeleted(self, ports):
        endpoint = self._get_endpoint('router_port')

        def resource_deleted():
            cvx_data = self.cvx.endpoint_data[endpoint]
            return all(port not in cvx_data for port in ports)
        common_utils.wait_until_true(resource_deleted)



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from oslo_config import cfg

from neutron.tests import base
from neutron.tests.unit import testlib_api

from networking_arista.common import exceptions as arista_exc
from networking_arista.ml2.rpc import arista_eapi
from networking_arista.tests.unit import utils


def setup_valid_config():
    utils.setup_arista_wrapper_config(cfg)


class AristaRPCWrapperInvalidConfigTestCase(base.BaseTestCase):
    """Negative test cases to test the Arista Driver configuration."""

    def setUp(self):
        super(AristaRPCWrapperInvalidConfigTestCase, self).setUp()
        self.setup_invalid_config()  # Invalid config, required options not set

    def setup_invalid_config(self):
        utils.setup_arista_wrapper_config(cfg, host='', user='')

    def test_raises_exception_on_wrong_configuration(self):
        self.assertRaises(arista_exc.AristaConfigError,
                          arista_eapi.AristaRPCWrapperEapi)


class NegativeRPCWrapperTestCase(testlib_api.SqlTestCase):
    """Negative test cases to test the RPC between Arista Driver and EOS."""

    def setUp(self):
        super(NegativeRPCWrapperTestCase, self).setUp()
        setup_valid_config()

    def test_exception_is_raised_on_json_server_error(self):
        drv = arista_eapi.AristaRPCWrapperEapi()

        drv.api_request = mock.MagicMock(
            side_effect=Exception('server error')
        )
        with mock.patch.object(arista_eapi.LOG, 'error') as log_err:
            self.assertRaises(arista_exc.AristaRpcError,
                              drv._run_openstack_cmds, [])
            log_err.assert_called_once_with(mock.ANY)



# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import json
import re

from oslo_log import log as logging

from neutron_lib import constants as n_const
from neutron_lib.db import api as db_api
from neutron_lib.services.trunk import constants as t_const

from neutron.db.models import l3 as l3_models
from neutron.db.models import l3ha as l3ha_models
from neutron.db.models import segment as segment_models
from neutron.db import models_v2
from neutron.plugins.ml2 import models as ml2_models
from neutron.services.trunk import models as t_models

from networking_arista.common import config  # noqa
from networking_arista.common import constants as a_const
from networking_arista.ml2 import arista_resources as resources

LOG = logging.getLogger(__name__)


def setup_arista_wrapper_config(cfg, host='host', user='user'):
    cfg.CONF.set_override('eapi_host', host, "ml2_arista")
    cfg.CONF.set_override('eapi_username', user, "ml2_arista")
    cfg.CONF.set_override('sync_interval', 1, "ml2_arista")
    cfg.CONF.set_override('conn_timeout', 20, "ml2_arista")
    cfg.CONF.set_override('switch_info', ['switch1:user:pass'], "ml2_arista")


class MockCvx(object):

    def __init__(self, region):
        whitelist = ['AristaResourcesBase',
                     'PortResourcesBase',
                     'AttributeFormatter']
        self.endpoint_data = {}
        self.endpoint_to_id = {}
        self.endpoint_to_class = {}
        self.region = region
        for cls in resources.__dict__.values():
            if (isinstance(cls, type) and
                    cls.__module__ == resources.__name__ and
                    cls.__name__ not in whitelist):
                region_endpoint = cls.endpoint % {'region': region}
                self.endpoint_data[region_endpoint] = {}
                self.endpoint_to_id[region_endpoint] = cls.id_key
                self.endpoint_to_class[region_endpoint] = cls

    def send_api_request(self, endpoint, request_type, data=None):
        if request_type != 'GET':
            LOG.debug("%(type)s %(endpoint)s %(data)s", {'type': request_type,
                                                         'endpoint': endpoint,
                                                         'data': data})
        if request_type == 'POST':
            for resource in data:
                endpoint_class = self.endpoint_to_class[endpoint]
                for key in endpoint_class.get_resource_ids(resource):
                    self.endpoint_data[endpoint][key] = resource
        elif request_type == 'GET':
            return self.endpoint_data[endpoint].values()
        elif request_type == 'DELETE':
            for resource in data:
                endpoint_class = self.endpoint_to_class[endpoint]
                for key in endpoint_class.get_resource_ids(resource):
                    try:
                        del self.endpoint_data[endpoint][key]
                    except KeyError:
                        pass

    def get_cvx_uuid(self):
        return None

    def sync_start(self):
        return True

    def sync_end(self):
        return True


class MockSwitch(object):

    def __init__(self):
        self._commands = []
        self._acl_rules = dict()
        self._bindings = defaultdict(lambda: defaultdict(set))
        self._vrfs = dict()
        self._svis = dict()
        self._vlans = dict()
        self._route = dict()
        self._acl_mode_re = re.compile('^(?P<delete>no )?ip access-list '
                                       '(?P<acl>\S+)(?P<dyn> dynamic)?$')
        self._interface_mode_re = re.compile(
            '^(?P<delete>no )?interface (?P<intf>.+)$')
        self._access_group_re = re.compile(
            '^(?P<delete>no )?ip access-group (?P<acl>\S+) (?P<dir>\S+)$')
        self._vrf_mode_re = re.compile(
            '^(?P<delete>no )?vrf instance (?P<name>\S+)$')
        self._vlan_re = re.compile('^(?P<delete>no )?vlan (?P<vlan>\d+)$')
        self._ip_address_re = re.compile(
            '^ip address (?P<ip>[\d.]+)/(?P<mask>\d+)$')
        self._vip_re = re.compile('^ip virtual-router address (?P<ip>[\d.]+)$')
        self._vrf_route_re = re.compile(
            '^(?P<delete>no )?ip route vrf (?P<vrf>\S+) '
            '(?P<network>[\d.]+/\d+) (?P<next_hop>[\d.]+)$')
        self._svi_vrf_re = re.compile('^vrf forwarding (?P<vrf>\S+)$')
        self._rd_re = re.compile('^rd (?P<rd>\S+)$')
        self._varp_mac_re = re.compile(
            '^ip virtual-router mac-address (?P<varp_mac>\S+)$')
        self._mode = None

    def execute(self, commands, commands_to_log=None, keep_alive=True):
        ret = []
        for command in commands:
            if command == 'show ip access-lists dynamic':
                acls = {'aclList': []}
                for name, rules in self._acl_rules.items():
                    acls['aclList'].append(
                        {'name': name,
                         'sequence': [{'text': rule} for rule in rules]})
                ret.append(acls)
            elif command == 'show ip access-lists summary dynamic':
                bindings = {'aclList': []}
                for acl, binding in self._bindings.items():
                    bindings['aclList'].append(
                        {'name': acl,
                         'configuredIngressIntfs': [
                             {'name': intf} for intf in
                             binding.get(a_const.INGRESS_DIRECTION, [])],
                         'configuredEgressIntfs': [
                             {'name': intf} for intf in
                             binding.get(a_const.EGRESS_DIRECTION, [])]})
                ret.append(bindings)
            elif command == 'show vlan':
                vlans = {'vlans': {}}
                for vlan, info in self._vlans.items():
                    vlans['vlans'][str(vlan)] = {'dynamic': info['dynamic']}
                ret.append(vlans)
            elif command == 'show ip interface':
                svis = {'interfaces': {}}
                for intf, svi in self._svis.items():
                    svis['interfaces']['Vlan%s' % intf.strip('vlan ')] = {
                        'interfaceAddress': [
                            {'primaryIp': {'maskLen': svi['mask'],
                                           'address': svi['ip']}}]}
                ret.append(svis)
            elif command == 'show vrf':
                vrfs = {'vrfs': {}}
                for vrf_name, vrf in self._vrfs.items():
                    vrfs['vrfs'][vrf_name] = {'interfaces': vrf['svis'],
                                              'routeDistinguisher': vrf['rd']}
                ret.append(vrfs)
            elif command == 'show ip route vrf all':
                vrfs = {'vrfs': {}}
                for vrf_name, vrf in self._vrfs.items():
                    vrfs['vrfs'][vrf_name] = {'routes': {}}
                    vrf_routes = vrf.get('routes')
                    if not vrf_routes:
                        continue
                    for route, vrf_route in vrf['routes'].items():
                        vrf_routes[route] = {
                            "kernelProgrammed": True,
                            "directlyConnected": False,
                            "routeAction": "forward",
                            "routeLeaked": False,
                            "vias": [
                                {
                                    "interface": "Management1",
                                    "nexthopAddr": vrf_route
                                }
                            ],
                            "metric": 0,
                            "hardwareProgrammed": True,
                            "routeType": "static",
                            "preference": 1
                        }
                ret.append(vrfs)
            elif command == 'show version':
                ret.append({'version': '4.22.0F'})
            elif command == 'enable':
                ret.append({})
            elif 'show' in command:
                pass
            elif 'ip access-list' in command:
                acl_match = self._acl_mode_re.match(command)
                acl = acl_match.group('acl')
                delete = acl_match.group('delete')
                if delete:
                    del self._acl_rules[acl]
                else:
                    assert acl_match.group('dyn'), "All ACLs must be dynamic"
                    self._mode = ('acl', acl)
                    if not self._acl_rules.get(acl):
                        self._acl_rules[acl] = list()
            elif 'interface' in command:
                intf_match = self._interface_mode_re.match(command)
                intf = intf_match.group('intf')
                if intf_match.group('delete'):
                    del self._svis[intf]
                else:
                    if 'vlan' in intf:
                        self._svis[intf] = {'ip': '',
                                            'mask': '',
                                            'vip': ''}
                    self._mode = ('interface', intf)
            elif 'vrf instance' in command:
                vrf_match = self._vrf_mode_re.match(command)
                delete = vrf_match.group('delete')
                vrf_name = vrf_match.group('name')
                if delete:
                    del self._vrfs[vrf_name]
                else:
                    self._vrfs[vrf_name] = {'svis': [], 'routes': {}}
                    self._mode = ('vrf', vrf_name)
            elif 'vlan' in command:
                self._parse_vlan(command)
            elif 'ip route vrf' in command:
                self._parse_ip_route_vrf(command)
            elif command == 'exit':
                self._mode = None
            else:
                if self._mode:
                    if self._mode[0] == 'acl':
                        if command.startswith('no '):
                            self._acl_rules[self._mode[1]].remove(command[3:])
                        else:
                            self._acl_rules[self._mode[1]].append(command)
                    elif self._mode[0] == 'interface':
                        acl_match = self._access_group_re.match(command)
                        if acl_match:
                            self._parse_acl(acl_match)
                        else:
                            self._parse_svi(command)
                    elif self._mode[0] == 'vrf':
                        self._parse_vrf(command)
            self._commands.append(command)
        return ret

    def _parse_acl(self, acl_match):
        delete = acl_match.group('delete')
        acl = acl_match.group('acl')
        direction = acl_match.group('dir')
        if delete:
            self._bindings[acl][direction].remove(
                self._mode[1])
        else:
            # Delete the old binding for this intf if nec.
            for binding in self._bindings.values():
                if self._mode[1] in binding[direction]:
                    binding[direction].remove(self._mode[1])
            self._bindings[acl][direction].add(self._mode[1])

    def _parse_svi(self, command):
        ip_addr_match = self._ip_address_re.match(command)
        if ip_addr_match:
            self._svis[self._mode[1]]['ip'] = ip_addr_match.group('ip')
            self._svis[self._mode[1]]['mask'] = ip_addr_match.group('mask')
        vip_match = self._vip_re.match(command)
        if vip_match:
            self._svis[self._mode[1]]['vip'] = vip_match.group('ip')
        vrf_match = self._svi_vrf_re.match(command)
        if vrf_match:
            self._vrfs[vrf_match.group('vrf')]['svis'].append(self._mode[1])

    def _parse_vrf(self, command):
        rd_match = self._rd_re.match(command)
        if rd_match:
            self._vrfs[self._mode[1]]['rd'] = rd_match.group('rd')
        varp_mac_match = self._varp_mac_re.match(command)
        if varp_mac_match:
            pass

    def _parse_ip_route_vrf(self, command):
        vrf_route_match = self._vrf_route_re.match(command)
        if vrf_route_match:
            delete = vrf_route_match.group('delete')
            vrf = vrf_route_match.group('vrf')
            if vrf not in self._vrfs:
                assert False
            vrf_dict = self._vrfs[vrf]
            vrf_routes = vrf_dict['routes']
            if delete:
                del self._vrfs[vrf]['routes'][vrf_route_match.group('network')]
            else:
                vrf_routes[vrf_route_match.group('network')] = \
                    vrf_route_match.group('next_hop')

    def _parse_vlan(self, command):
        vlan_match = self._vlan_re.match(command)
        delete = vlan_match.group('delete')
        vlan = vlan_match.group('vlan')
        if delete:
            del self._vlans[vlan]
        else:
            self._vlans[vlan] = {'dynamic': False}

    @property
    def received_commands(self):
        return self._commands

    def assert_command_not_received(self, unexpected_cmd):
        for cmd in self._commands:
            assert unexpected_cmd not in cmd

    def clear_received_commands(self):
        self._commands = []

    def reset_switch(self):
        self._commands = []
        self._acl_rules = dict()
        self._bindings = defaultdict(lambda: defaultdict(set))
        self._vrfs = dict()
        self._svis = dict()
        self._vlans = dict()


# Network utils #


def create_networks(networks):
    session = db_api.get_writer_session()
    with session.begin():
        for network in networks:
            session.add(models_v2.Network(**network))


def delete_network(network_id):
    session = db_api.get_writer_session()
    with session.begin():
        network_model = models_v2.Network
        session.query(network_model).filter(
            network_model.id == network_id).delete()


def delete_networks_for_tenant(tenant_id):
    session = db_api.get_writer_session()
    with session.begin():
        network_model = models_v2.Network
        networks = session.query(network_model).filter(
            network_model.project_id == tenant_id).all()
        for network in networks:
            delete_ports_on_network(network.id)
            session.delete(network)


# Segment utils #


def create_segments(segments):
    session = db_api.get_writer_session()
    with session.begin():
        for segment in segments:
            session.add(segment_models.NetworkSegment(**segment))


def delete_segment(segment_id):
    session = db_api.get_writer_session()
    with session.begin():
        segment_model = segment_models.NetworkSegment
        session.query(segment_model).filter(
            segment_model.id == segment_id).delete()


def delete_segments_for_network(network_id):
    session = db_api.get_writer_session()
    with session.begin():
        segment_model = segment_models.NetworkSegment
        session.query(segment_model).filter(
            segment_model.network_id == network_id).delete()


def delete_segments_for_tenant(tenant_id):
    session = db_api.get_writer_session()
    network_model = models_v2.Network
    segment_model = segment_models.NetworkSegment
    with session.begin():
        networks = session.query(network_model).filter(
            network_model.project_id == tenant_id).all()
        for network in networks:
            session.query(segment_model).filter(
                segment_model.network_id == network.id).delete()


# Port utils #


def create_ports(ports):
    session = db_api.get_writer_session()
    with session.begin():
        for port in ports:
            binding_levels = port.pop('binding_levels', [])
            binding = port.pop('binding', {})
            session.add(models_v2.Port(**port))
            if binding:
                binding['port_id'] = port['id']
                if binding['vif_type'] == 'distributed':
                    distributed_binding = binding.copy()
                    distributed_binding['status'] = 'ACTIVE'
                    for host in binding['host']:
                        distributed_binding['host'] = host
                        session.add(
                            ml2_models.DistributedPortBinding(
                                **distributed_binding))
                else:
                    session.add(ml2_models.PortBinding(**binding))
            for binding_level in binding_levels:
                binding_level['port_id'] = port['id']
                session.add(ml2_models.PortBindingLevel(**binding_level))


def delete_port(port_id):
    session = db_api.get_writer_session()
    with session.begin():
        port_model = models_v2.Port
        session.query(port_model).filter(
            port_model.id == port_id).delete()


def delete_ports_on_network(network_id):
    session = db_api.get_writer_session()
    with session.begin():
        port_model = models_v2.Port
        session.query(port_model).filter(
            port_model.network_id == network_id).delete()


def delete_ports_for_instance(instance_id):
    session = db_api.get_writer_session()
    with session.begin():
        port_model = models_v2.Port
        session.query(port_model).filter(
            port_model.device_id == instance_id).delete()


def delete_ports_for_tenant(tenant_id):
    session = db_api.get_writer_session()
    with session.begin():
        port_model = models_v2.Port
        session.query(port_model).filter(
            port_model.project_id == tenant_id).delete()


# Port binding utils #


def delete_port_binding(port_id, host):
    session = db_api.get_writer_session()
    with session.begin():
        # We cannot do any bulk deletes here because every delete bumps the
        # revision number of the Port
        pbl_model = ml2_models.PortBindingLevel
        levels = (session.query(pbl_model)
                  .filter(pbl_model.port_id == port_id,
                          pbl_model.host == host))
        for level in levels:
            session.delete(level)
        pb_model = ml2_models.PortBinding
        bindings = (session.query(pb_model)
                    .filter(pb_model.port_id == port_id,
                            pb_model.host == host))
        for binding in bindings:
            session.delete(binding)
        dpb_model = ml2_models.DistributedPortBinding
        bindings = (session.query(dpb_model)
                    .filter(dpb_model.port_id == port_id,
                            dpb_model.host == host))
        for binding in bindings:
            session.delete(binding)


def remove_switch_binding(port_id, switch_id, intf_id):
    session = db_api.get_writer_session()
    with session.begin():
        pb_model = ml2_models.PortBinding
        binding = (session.query(pb_model)
                   .filter(pb_model.port_id == port_id).first())
        profile = json.loads(binding.profile)
        lli = profile['local_link_information']
        for idx, link in enumerate(lli):
            if link['switch_id'] == switch_id and link['port_id'] == intf_id:
                lli.pop(idx)
                break
        binding.profile = json.dumps(profile)
    if len(lli) == 0:
        delete_port_binding(port_id, binding.host)


# Trunk utils #


def create_trunks(trunks):
    session = db_api.get_writer_session()
    with session.begin():
        for trunk in trunks:
            session.add(t_models.Trunk(**trunk))


def create_subports(subports):
    session = db_api.get_writer_session()
    with session.begin():
        for subport in subports:
            session.add(t_models.SubPort(**subport))


# L3 HA Router utils #

def create_ha_routers(routers):
    session = db_api.get_writer_session()
    with session.begin():
        for router in routers:
            session.add(l3_models.Router(**router))


def create_ha_router_networks(networks):
    session = db_api.get_writer_session()
    with session.begin():
        for network in networks:
            session.add(l3ha_models.L3HARouterNetwork(**network))


def delete_ha_router_for_tenant(tenant_id):
    session = db_api.get_writer_session()
    with session.begin():
        router_model = l3_models.Router
        routers = session.query(router_model).filter(
            router_model.project_id == tenant_id).all()
        for router in routers:
            session.delete(router)


def setup_scenario():
    # Create networks
    regular_network = {'id': 'n1',
                       'project_id': 't1',
                       'name': 'regular',
                       'admin_state_up': True,
                       'rbac_entries': []}
    hpb_network = {'id': 'n2',
                   'project_id': 't2',
                   'name': 'hpb',
                   'admin_state_up': True,
                   'rbac_entries': []}

    ha_network = {'id': 'HA network',
                  'project_id': '',
                  'name': 'l3 ha',
                  'admin_state_up': True,
                  'rbac_entries': []}

    # Create segments
    flat_segment = {'id': 'sError',
                    'network_id': 'n1',
                    'is_dynamic': False,
                    'network_type': 'flat'}
    regular_segment = {'id': 's1',
                       'network_id': 'n1',
                       'is_dynamic': False,
                       'segmentation_id': 11,
                       'network_type': 'vlan',
                       'physical_network': 'default'}
    fabric_segment = {'id': 's2',
                      'network_id': 'n2',
                      'is_dynamic': False,
                      'segmentation_id': 20001,
                      'network_type': 'vxlan',
                      'physical_network': None}
    dynamic_segment1 = {'id': 's3',
                        'network_id': 'n2',
                        'is_dynamic': True,
                        'segmentation_id': 21,
                        'network_type': 'vlan',
                        'physical_network': 'switch1'}
    dynamic_segment2 = {'id': 's4',
                        'network_id': 'n2',
                        'is_dynamic': True,
                        'segmentation_id': 31,
                        'network_type': 'vlan',
                        'physical_network': 'switch2'}
    ha_segment = {'id': 's5',
                  'network_id': 'HA network',
                  'is_dynamic': False,
                  'segmentation_id': 33,
                  'network_type': 'vlan',
                  'physical_network': 'default'}
    ha_fabric_segment = {'id': 's6',
                         'network_id': 'HA network',
                         'is_dynamic': False,
                         'segmentation_id': 20010,
                         'network_type': 'vxlan',
                         'physical_network': None}
    ha_dynamic_segment = {'1': {'id': 's7',
                                'network_id': 'HA network',
                                'is_dynamic': True,
                                'segmentation_id': 700,
                                'network_type': 'vlan',
                                'physical_network': 'switch1'},
                          '2': {'id': 's8',
                                'network_id': 'HA network',
                                'is_dynamic': True,
                                'segmentation_id': 800,
                                'network_type': 'vlan',
                                'physical_network': 'switch2'}}

    # Create ports
    port_ctr = 0
    ports = list()
    trunk_ctr = 0
    trunks = list()
    subports = list()
    instance_types = [(n_const.DEVICE_OWNER_DHCP, 'normal'),
                      (n_const.DEVICE_OWNER_DVR_INTERFACE, 'normal'),
                      (n_const.DEVICE_OWNER_COMPUTE_PREFIX, 'normal'),
                      (n_const.DEVICE_OWNER_COMPUTE_PREFIX, 'baremetal'),
                      (n_const.DEVICE_OWNER_BAREMETAL_PREFIX, 'baremetal'),
                      (n_const.DEVICE_OWNER_ROUTER_HA_INTF, 'normal'),
                      (n_const.DEVICE_OWNER_ROUTER_INTF, 'normal')]
    for device_owner, vnic_type in instance_types:
        vif_type = 'ovs'
        regular_host = 'host1'
        regular_binding_levels = [
            {'host': 'host1',
             'segment_id': regular_segment['id'],
             'level': 0,
             'driver': 'arista'}]
        hpb_binding_levels = [
            {'host': 'host2',
             'segment_id': fabric_segment['id'],
             'level': 0,
             'driver': 'arista'},
            {'host': 'host2',
             'segment_id': dynamic_segment1['id'],
             'level': 1,
             'driver': 'arista'}]
        hpb_host = 'host2'
        binding_profile = ''
        if vnic_type == 'baremetal':
            binding_profile = ('{"local_link_information": ' +
                               '[{"switch_id": "00:11:22:33:44:55", ' +
                               '"port_id": "Ethernet1"}, ' +
                               '{"switch_id": "00:11:22:33:44:55", ' +
                               '"port_id": "Ethernet2"}, ' +
                               '{"switch_id": "55:44:33:22:11:00", ' +
                               '"port_id": "Ethernet1/1"}, ' +
                               '{"switch_id": "55:44:33:22:11:00", ' +
                               '"port_id": "Ethernet1/2"}]}')

        if device_owner == n_const.DEVICE_OWNER_DVR_INTERFACE:
            vif_type = 'distributed'
            regular_host = ['host1', 'host2']
            regular_binding_levels = [
                {'host': 'host1',
                 'segment_id': regular_segment['id'],
                 'level': 0,
                 'driver': 'arista'},
                {'host': 'host2',
                 'segment_id': regular_segment['id'],
                 'level': 0,
                 'driver': 'arista'}]
            hpb_binding_levels = [
                {'host': 'host1',
                 'segment_id': fabric_segment['id'],
                 'level': 0,
                 'driver': 'arista'},
                {'host': 'host1',
                 'segment_id': dynamic_segment1['id'],
                 'level': 1,
                 'driver': 'arista'},
                {'host': 'host2',
                 'segment_id': fabric_segment['id'],
                 'level': 0,
                 'driver': 'arista'},
                {'host': 'host2',
                 'segment_id': dynamic_segment2['id'],
                 'level': 1,
                 'driver': 'arista'}]
            hpb_host = ['host1', 'host2']

        port_ctr += 1
        regular_port = {'admin_state_up': True,
                        'status': 'ACTIVE',
                        'device_id': '%s%s1' % (device_owner, vnic_type),
                        'device_owner': device_owner,
                        'binding': {'host': regular_host,
                                    'vif_type': vif_type,
                                    'vnic_type': vnic_type,
                                    'profile': binding_profile},
                        'tenant_id': 't1',
                        'id': 'p%d' % port_ctr,
                        'network_id': regular_network['id'],
                        'mac_address': '00:00:00:00:00:%02x' % port_ctr,
                        'name': 'regular_port',
                        'binding_levels': regular_binding_levels}
        port_ctr += 1
        hpb_port = {'admin_state_up': True,
                    'status': 'ACTIVE',
                    'device_id': '%s%s2' % (device_owner, vnic_type),
                    'device_owner': device_owner,
                    'binding': {'host': hpb_host,
                                'vif_type': vif_type,
                                'vnic_type': vnic_type,
                                'profile': binding_profile},
                    'tenant_id': 't2',
                    'id': 'p%d' % port_ctr,
                    'network_id': hpb_network['id'],
                    'mac_address': '00:00:00:00:00:%02x' % port_ctr,
                    'name': 'hpb_port',
                    'binding_levels': hpb_binding_levels}
        if device_owner != n_const.DEVICE_OWNER_ROUTER_HA_INTF:
            ports.extend([regular_port, hpb_port])
        if device_owner == n_const.DEVICE_OWNER_COMPUTE_PREFIX:
            port_ctr += 1
            trunk_subport = {'admin_state_up': True,
                             'status': 'ACTIVE',
                             'device_id': '%s%s1' % (device_owner, vnic_type),
                             'device_owner': t_const.TRUNK_SUBPORT_OWNER,
                             'binding': {'host': regular_host,
                                         'vif_type': vif_type,
                                         'vnic_type': vnic_type,
                                         'profile': binding_profile},
                             'tenant_id': 't1',
                             'id': 'p%d' % port_ctr,
                             'network_id': regular_network['id'],
                             'mac_address': '10:00:00:00:00:%02x' % port_ctr,
                             'name': 'trunk_subport',
                             'binding_levels': regular_binding_levels}
            ports.extend([trunk_subport])
            trunk = {'id': 't%d' % trunk_ctr,
                     'port_id': regular_port['id']}
            subport = {'port_id': trunk_subport['id'],
                       'trunk_id': trunk['id'],
                       'segmentation_type': 'vlan',
                       'segmentation_id': 100}
            trunk_ctr += 1
            trunks.append(trunk)
            subports.append(subport)
        if device_owner == n_const.DEVICE_OWNER_ROUTER_HA_INTF:
            vif_type = 'ovs'
            ha_router_ports = [{'admin_state_up': True,
                                'status': 'ACTIVE',
                                'device_id': '%s%s' % (device_owner,
                                                       vnic_type),
                                'device_owner': device_owner,
                                'binding': {'host': 'host%d' % p,
                                            'vif_type': vif_type,
                                            'vnic_type': vnic_type,
                                            'profile': binding_profile},
                                'tenant_id': '',
                                'id': 'uuid-ha-%d' % p,
                                'network_id': ha_network['id'],
                                'mac_address': '00:00:00:00:00:%02x' % p,
                                'name': 'regular_port',
                                'binding_levels': [
                                    {'host': 'host%d' % p,
                                     'segment_id': ha_segment['id'],
                                     'level': 0,
                                     'driver': 'openvswitch'}]}
                               for p in range(1, 3)]
            ports.extend(ha_router_ports)
            ha_router_hpb_ports = [
                {'admin_state_up': True,
                 'status': 'ACTIVE',
                 'device_id': '%s%s' % (device_owner,
                                        vnic_type),
                 'device_owner': device_owner,
                 'binding': {'host': 'host%d' % p,
                             'vif_type': vif_type,
                             'vnic_type': vnic_type,
                             'profile': binding_profile},
                 'tenant_id': '',
                 'id': 'uuid-hpb-ha-%d' % p,
                 'network_id': ha_network['id'],
                 'mac_address': '00:00:00:00:01:%02x' % p,
                 'name': 'ha_router_hpb_port',
                 'binding_levels': [
                     {'host': 'host%d' % p,
                      'segment_id': ha_fabric_segment['id'],
                      'level': 0,
                      'driver': 'arista'},
                     {'host': 'host%d' % p,
                      'segment_id': ha_dynamic_segment['%d' % p]['id'],
                      'level': 1,
                      'driver': 'openvswitch'}]} for p in range(1, 3)]
            ports.extend(ha_router_hpb_ports)

    routers = [
        {'project_id': 'ha-router-project',
         'id': n_const.DEVICE_OWNER_ROUTER_HA_INTF + 'normal',
         'name': 'test-router'}]
    create_ha_routers(routers)
    create_networks([regular_network, hpb_network, ha_network])
    ha_router_networks = [
        {'project_id': 'ha-router-project',
         'network_id': ha_network['id']}]
    create_ha_router_networks(ha_router_networks)
    create_segments([regular_segment, fabric_segment, flat_segment,
                     dynamic_segment1, dynamic_segment2, ha_segment,
                     ha_fabric_segment, ha_dynamic_segment['1'],
                     ha_dynamic_segment['2']])
    create_ports(ports)
    create_trunks(trunks)
    create_subports(subports)



# Copyright (c) 2018 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib import constants as n_const

from networking_arista.tests.unit.ml2.security_groups import sg_test_base


class SecurityGroupCallbacksTestCase(sg_test_base.SecurityGroupTestBase):

    def test_create_security_group(self):
        sec_group = {'security_group':
                     {'name': 'sg1',
                      'tenant_id': 't1',
                      'description': ''}}
        grp = self.plugin.create_security_group(self.context, sec_group,
                                                default_sg=True)
        expected_eapi_commands = [
            'enable',
            'configure',
            'ip access-list SG-INGRESS-%s dynamic' % grp['id'],
            'exit',
            'ip access-list SG-EGRESS-%s dynamic' % grp['id'],
            'exit',
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_delete_security_group(self):
        sec_group = {'security_group':
                     {'name': 'sg1',
                      'tenant_id': 't1',
                      'description': ''}}
        grp = self.plugin.create_security_group(self.context, sec_group,
                                                default_sg=True)
        for switch in self.switches.values():
            switch.clear_received_commands()
        self.plugin.delete_security_group(self.context, grp['id'])
        expected_eapi_commands = [
            'enable',
            'configure',
            'no ip access-list SG-INGRESS-%s' % grp['id'],
            'no ip access-list SG-EGRESS-%s' % grp['id'],
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_create_security_group_rule_ingress(self):
        direction = 'ingress'
        proto = 'tcp'
        cidr = '10.0.0.0/24'
        grp_id, _ = self.create_sg_rule(direction, proto, cidr)
        expected_eapi_commands = [
            'enable',
            'configure',
            'ip access-list SG-INGRESS-%s dynamic' % grp_id,
            'permit %s %s any' % (proto, cidr),
            'exit',
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_create_security_group_rule_egress(self):
        direction = 'egress'
        proto = 'tcp'
        cidr = '10.0.0.0/24'
        grp_id, _ = self.create_sg_rule(direction, proto, cidr)
        expected_eapi_commands = [
            'enable',
            'configure',
            'ip access-list SG-EGRESS-%s dynamic' % grp_id,
            'permit %s any %s' % (proto, cidr),
            'exit',
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_create_security_group_rule_tcp(self):
        direction = 'egress'
        proto = 'tcp'
        cidr = '10.0.0.0/24'
        grp_id, _ = self.create_sg_rule(direction, proto, cidr)
        expected_eapi_commands = [
            'enable',
            'configure',
            'ip access-list SG-EGRESS-%s dynamic' % grp_id,
            'permit %s any %s' % (proto, cidr),
            'exit',
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_create_security_group_rule_udp(self):
        direction = 'egress'
        proto = 'udp'
        cidr = '10.0.0.0/24'
        grp_id, _ = self.create_sg_rule(direction, proto, cidr)
        expected_eapi_commands = [
            'enable',
            'configure',
            'ip access-list SG-EGRESS-%s dynamic' % grp_id,
            'permit %s any %s' % (proto, cidr),
            'exit',
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_create_security_group_rule_port_range(self):
        direction = 'egress'
        proto = 'tcp'
        cidr = '10.0.0.0/24'
        range_min = 100
        range_max = 200
        grp_id, _ = self.create_sg_rule(direction, proto, cidr,
                                        range_min, range_max)
        expected_eapi_commands = [
            'enable',
            'configure',
            'ip access-list SG-EGRESS-%s dynamic' % grp_id,
            'permit %s any %s range %s %s' % (proto, cidr,
                                              range_min, range_max),
            'exit',
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_create_security_group_rule_basic_icmp(self):
        direction = 'egress'
        proto = 'icmp'
        cidr = '10.0.0.0/24'
        grp_id, _ = self.create_sg_rule(direction, proto, cidr)
        expected_eapi_commands = [
            'enable',
            'configure',
            'ip access-list SG-EGRESS-%s dynamic' % grp_id,
            'permit %s any %s' % (proto, cidr),
            'exit',
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_create_security_group_rule_icmp_type(self):
        direction = 'egress'
        proto = 'icmp'
        cidr = '10.0.0.0/24'
        message_type = 10
        grp_id, _ = self.create_sg_rule(direction, proto, cidr, message_type)
        expected_eapi_commands = [
            'enable',
            'configure',
            'ip access-list SG-EGRESS-%s dynamic' % grp_id,
            'permit %s any %s %s' % (proto, cidr, message_type),
            'exit',
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_create_security_group_rule_icmp_code(self):
        direction = 'egress'
        proto = 'icmp'
        cidr = '10.0.0.0/24'
        message_type = 10
        message_code = 100
        grp_id, _ = self.create_sg_rule(direction, proto, cidr, message_type,
                                        message_code)
        expected_eapi_commands = [
            'enable',
            'configure',
            'ip access-list SG-EGRESS-%s dynamic' % grp_id,
            'permit %s any %s %s %s' % (proto, cidr, message_type,
                                        message_code),
            'exit',
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_create_security_group_no_ip(self):
        direction = 'egress'
        proto = 'tcp'
        cidr = None
        grp_id, _ = self.create_sg_rule(direction, proto, cidr)
        expected_eapi_commands = [
            'enable',
            'configure',
            'ip access-list SG-EGRESS-%s dynamic' % grp_id,
            'permit %s any any' % proto,
            'exit',
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_create_security_group_ipv6(self):
        direction = 'egress'
        proto = 'tcp'
        cidr = None
        ethertype = 'IPv6'
        grp_id, _ = self.create_sg_rule(direction, proto, cidr,
                                        ethertype=ethertype)
        for switch in self.switches.values():
            self.assertEqual([], switch.received_commands)

    def test_delete_security_group_rule(self):
        direction = 'egress'
        proto = 'tcp'
        cidr = '10.0.0.0/24'
        grp_id, rule = self.create_sg_rule(direction, proto, cidr)
        for switch in self.switches.values():
            switch.clear_received_commands()
        self.plugin.delete_security_group_rule(self.context, rule['id'])
        expected_eapi_commands = [
            'enable',
            'configure',
            'ip access-list SG-EGRESS-%s dynamic' % grp_id,
            'no permit %s any %s' % (proto, cidr),
            'exit',
            'exit']
        for switch in self.switches.values():
            self.assertEqual(expected_eapi_commands,
                             switch.received_commands)

    def test_apply_security_group(self):
        switch_port = 'Ethernet1'
        switch_id = '11:22:33:44:55'
        switch_info = 'TOR1'
        grp_id, rule = self.create_sg_rule('egress', 'tcp', '10.0.0.0/24')
        for switch in self.switches.values():
            switch.clear_received_commands()
        net_dict = {'network': {'name': 'net',
                                'tenant_id': 't1',
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)
        port_dict = {'name': 'port1',
                     'tenant_id': 't1',
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': 'bm1',
                     'device_owner': n_const.DEVICE_OWNER_BAREMETAL_PREFIX,
                     'binding:host_id': 'bm-host',
                     'binding:profile': {'local_link_information': [
                         {'switch_id': switch_id,
                          'port_id': switch_port,
                          'switch_info': switch_info}]},
                     'binding:vnic_type': 'baremetal',
                     'security_groups': [grp_id]}
        port, _ = self.create_port(port_dict)
        expected_eapi_commands = [
            'show interfaces',
            'enable',
            'configure',
            'interface %s' % switch_port,
            'ip access-group SG-INGRESS-%s out' % grp_id,
            'ip access-group SG-EGRESS-%s in' % grp_id,
            'exit',
            'exit']
        # SGs are applied on binding and on status DOWN->UP,
        # so expect the commands twice
        expected_eapi_commands.extend(expected_eapi_commands)
        self.assertEqual(expected_eapi_commands,
                         self.switch1.received_commands)
        self.assertEqual([], self.switch2.received_commands)

    def test_apply_security_group_lag(self):
        switch_port = 'Ethernet1'
        port_channel = 'Port-Channel100'
        switch_id = '11:22:33:44:55'
        switch_info = 'TOR1'
        self.create_port_channel(switch_info, switch_port, port_channel)
        grp_id, rule = self.create_sg_rule('egress', 'tcp', '10.0.0.0/24')
        for switch in self.switches.values():
            switch.clear_received_commands()
        net_dict = {'network': {'name': 'net',
                                'tenant_id': 't1',
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)
        port_dict = {'name': 'port1',
                     'tenant_id': 't1',
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': 'bm1',
                     'device_owner': n_const.DEVICE_OWNER_BAREMETAL_PREFIX,
                     'binding:host_id': 'bm-host',
                     'binding:profile': {'local_link_information': [
                         {'switch_id': switch_id,
                          'port_id': switch_port,
                          'switch_info': switch_info}]},
                     'binding:vnic_type': 'baremetal',
                     'security_groups': [grp_id]}
        port, _ = self.create_port(port_dict)
        expected_eapi_commands = [
            'show interfaces',
            'enable',
            'configure',
            'interface %s' % port_channel,
            'ip access-group SG-INGRESS-%s out' % grp_id,
            'ip access-group SG-EGRESS-%s in' % grp_id,
            'exit',
            'exit']
        # SGs are applied on binding and on status DOWN->UP,
        # so expect the commands twice
        expected_eapi_commands.extend(expected_eapi_commands)
        self.assertEqual(expected_eapi_commands,
                         self.switch1.received_commands)
        self.assertEqual([], self.switch2.received_commands)

    def test_apply_security_group_mlag(self):
        switch_port = 'Ethernet1'
        port_channel = 'Port-Channel100'
        switch_id = '11:22:33:44:55'
        switch1_info = 'TOR1'
        switch2_info = 'TOR2'
        self.create_port_channel(switch1_info, switch_port, port_channel)
        self.create_port_channel(switch2_info, switch_port, port_channel)
        grp_id, rule = self.create_sg_rule('egress', 'tcp', '10.0.0.0/24')
        for switch in self.switches.values():
            switch.clear_received_commands()
        net_dict = {'network': {'name': 'net',
                                'tenant_id': 't1',
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)
        port_dict = {'name': 'port1',
                     'tenant_id': 't1',
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': 'bm1',
                     'device_owner': n_const.DEVICE_OWNER_BAREMETAL_PREFIX,
                     'binding:host_id': 'bm-host',
                     'binding:profile': {'local_link_information': [
                         {'switch_id': switch_id,
                          'port_id': switch_port,
                          'switch_info': switch1_info},
                         {'switch_id': switch_id,
                          'port_id': switch_port,
                          'switch_info': switch2_info}]},
                     'binding:vnic_type': 'baremetal',
                     'security_groups': [grp_id]}
        port, _ = self.create_port(port_dict)
        expected_eapi_commands = [
            'show interfaces',
            'enable',
            'configure',
            'interface %s' % port_channel,
            'ip access-group SG-INGRESS-%s out' % grp_id,
            'ip access-group SG-EGRESS-%s in' % grp_id,
            'exit',
            'exit']
        # SGs are applied on binding and on status DOWN->UP,
        # so expect the commands twice
        expected_eapi_commands.extend(expected_eapi_commands)
        self.assertEqual(expected_eapi_commands,
                         self.switch1.received_commands)
        self.assertEqual(expected_eapi_commands,
                         self.switch2.received_commands)

    def test_remove_security_group(self):
        switch_port = 'Ethernet1'
        switch_id = '11:22:33:44:55'
        switch_info = 'TOR1'
        grp_id, rule = self.create_sg_rule('egress', 'tcp', '10.0.0.0/24')
        net_dict = {'network': {'name': 'net',
                                'tenant_id': 't1',
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)
        port_dict = {'name': 'port1',
                     'tenant_id': 't1',
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': 'bm1',
                     'device_owner': n_const.DEVICE_OWNER_BAREMETAL_PREFIX,
                     'binding:host_id': 'bm-host',
                     'binding:profile': {'local_link_information': [
                         {'switch_id': switch_id,
                          'port_id': switch_port,
                          'switch_info': switch_info}]},
                     'binding:vnic_type': 'baremetal',
                     'security_groups': [grp_id]}
        port, _ = self.create_port(port_dict)
        for switch in self.switches.values():
            switch.clear_received_commands()
        self.delete_port(port['id'])
        expected_eapi_commands = [
            'show interfaces',
            'enable',
            'configure',
            'interface %s' % switch_port,
            'no ip access-group SG-INGRESS-%s out' % grp_id,
            'no ip access-group SG-EGRESS-%s in' % grp_id,
            'exit',
            'exit']
        self.assertEqual(expected_eapi_commands,
                         self.switch1.received_commands)
        self.assertEqual([], self.switch2.received_commands)

    def test_remove_security_group_lag(self):
        switch_port = 'Ethernet1'
        port_channel = 'Port-Channel100'
        switch_id = '11:22:33:44:55'
        switch_info = 'TOR1'
        self.create_port_channel(switch_info, switch_port, port_channel)
        grp_id, rule = self.create_sg_rule('egress', 'tcp', '10.0.0.0/24')
        net_dict = {'network': {'name': 'net',
                                'tenant_id': 't1',
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)
        port_dict = {'name': 'port1',
                     'tenant_id': 't1',
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': 'bm1',
                     'device_owner': n_const.DEVICE_OWNER_BAREMETAL_PREFIX,
                     'binding:host_id': 'bm-host',
                     'binding:profile': {'local_link_information': [
                         {'switch_id': switch_id,
                          'port_id': switch_port,
                          'switch_info': switch_info}]},
                     'binding:vnic_type': 'baremetal',
                     'security_groups': [grp_id]}
        port, _ = self.create_port(port_dict)
        for switch in self.switches.values():
            switch.clear_received_commands()
        self.delete_port(port['id'])
        expected_eapi_commands = [
            'show interfaces',
            'enable',
            'configure',
            'interface %s' % port_channel,
            'no ip access-group SG-INGRESS-%s out' % grp_id,
            'no ip access-group SG-EGRESS-%s in' % grp_id,
            'exit',
            'exit']
        self.assertEqual(expected_eapi_commands,
                         self.switch1.received_commands)
        self.assertEqual([], self.switch2.received_commands)

    def test_remove_security_group_mlag(self):
        switch_port = 'Ethernet1'
        port_channel = 'Port-Channel100'
        switch_id = '11:22:33:44:55'
        switch1_info = 'TOR1'
        switch2_info = 'TOR2'
        self.create_port_channel(switch1_info, switch_port, port_channel)
        self.create_port_channel(switch2_info, switch_port, port_channel)
        grp_id, rule = self.create_sg_rule('egress', 'tcp', '10.0.0.0/24')
        net_dict = {'network': {'name': 'net',
                                'tenant_id': 't1',
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)
        port_dict = {'name': 'port1',
                     'tenant_id': 't1',
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': 'bm1',
                     'device_owner': n_const.DEVICE_OWNER_BAREMETAL_PREFIX,
                     'binding:host_id': 'bm-host',
                     'binding:profile': {'local_link_information': [
                         {'switch_id': switch_id,
                          'port_id': switch_port,
                          'switch_info': switch1_info},
                         {'switch_id': switch_id,
                          'port_id': switch_port,
                          'switch_info': switch2_info}]},
                     'binding:vnic_type': 'baremetal',
                     'security_groups': [grp_id]}
        port, _ = self.create_port(port_dict)
        for switch in self.switches.values():
            switch.clear_received_commands()
        self.delete_port(port['id'])
        expected_eapi_commands = [
            'show interfaces',
            'enable',
            'configure',
            'interface %s' % port_channel,
            'no ip access-group SG-INGRESS-%s out' % grp_id,
            'no ip access-group SG-EGRESS-%s in' % grp_id,
            'exit',
            'exit']
        self.assertEqual(expected_eapi_commands,
                         self.switch1.received_commands)
        self.assertEqual(expected_eapi_commands,
                         self.switch2.received_commands)

    def test_apply_security_group_vm(self):
        grp_id, rule = self.create_sg_rule('egress', 'tcp', '10.0.0.0/24')
        net_dict = {'network': {'name': 'net',
                                'tenant_id': 't1',
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)
        port_dict = {'name': 'port1',
                     'tenant_id': 't1',
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': 'vm1',
                     'device_owner': n_const.DEVICE_OWNER_COMPUTE_PREFIX,
                     'binding:host_id': self.host1,
                     'binding:vnic_type': 'normal',
                     'security_groups': [grp_id]}
        for switch in self.switches.values():
            switch.clear_received_commands()
        self.create_port(port_dict)
        self.assertEqual([], self.switch1.received_commands)
        self.assertEqual([], self.switch2.received_commands)

    def test_apply_multiple_security_groups(self):
        switch_id = '00:11:22:33:44:55'
        switch_info = 'TOR1'
        switch_port = 'Ethernet1'
        proto = 'tcp'
        cidr = '10.0.0.0/24'
        grp1_id, _ = self.create_sg_rule('egress', proto, cidr)
        grp2_id, _ = self.create_sg_rule('ingress', proto, cidr,
                                         default=False)
        net_dict = {'network': {'name': 'net',
                                'tenant_id': 't1',
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)
        port_dict = {'name': 'port1',
                     'tenant_id': 't1',
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': 'bm1',
                     'device_owner': n_const.DEVICE_OWNER_BAREMETAL_PREFIX,
                     'binding:host_id': 'bm-host',
                     'binding:profile': {'local_link_information': [
                         {'switch_id': switch_id,
                          'port_id': switch_port,
                          'switch_info': switch_info}]},
                     'binding:vnic_type': 'baremetal',
                     'security_groups': [grp1_id, grp2_id]}
        for switch in self.switches.values():
            switch.clear_received_commands()
        port, _ = self.create_port(port_dict)
        self.assertEqual([], self.switch1.received_commands)
        self.assertEqual([], self.switch2.received_commands)



# Copyright (c) 2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import mock
from mock import patch
from neutron_lib.db import api as db_api
from oslo_config import cfg

from neutron.db.models.plugins.ml2 import vlanallocation
from neutron.tests.unit import testlib_api

from networking_arista.ml2.drivers.driver_helpers import VlanSyncService
from networking_arista.ml2.drivers.type_arista_vlan import AristaVlanTypeDriver
import networking_arista.tests.unit.ml2.utils as utils


EAPI_SEND_FUNC = ('networking_arista.ml2.rpc.arista_eapi.AristaRPCWrapperEapi'
                  '._send_eapi_req')


class AristaTypeDriverTest(testlib_api.SqlTestCase):

    def setUp(self):
        super(AristaTypeDriverTest, self).setUp()
        utils.setup_arista_wrapper_config(cfg)

    @patch(EAPI_SEND_FUNC)
    def test_initialize_type_driver(self, mock_send_eapi_req):
        type_driver = AristaVlanTypeDriver()
        type_driver.sync_service._force_sync = False
        type_driver.sync_service._vlan_assignment_uuid = {'uuid': 1}
        type_driver.sync_service._rpc = mock.MagicMock()
        rpc = type_driver.sync_service._rpc
        rpc.get_vlan_assignment_uuid.return_value = {'uuid': 1}
        type_driver.initialize()

        cmds = ['show openstack agent uuid',
                'show openstack instances',
                'show openstack agent uuid',
                'show openstack features']

        calls = [mock.call(cmds=[cmd], commands_to_log=[cmd])
                 for cmd in cmds]
        mock_send_eapi_req.assert_has_calls(calls)
        type_driver.timer.cancel()


class VlanSyncServiceTest(testlib_api.SqlTestCase):
    """Test that VLANs are synchronized between EOS and Neutron."""

    def _ensure_in_db(self, assigned, allocated, available):
        session = db_api.get_reader_session()
        with session.begin():
            vlans = session.query(vlanallocation.VlanAllocation).all()
            for vlan in vlans:
                self.assertIn(vlan.vlan_id, assigned)

                if vlan.vlan_id in available:
                    self.assertFalse(vlan.allocated)
                elif vlan.vlan_id in allocated:
                    self.assertTrue(vlan.allocated)

    def test_synchronization_test(self):
        rpc = mock.MagicMock()

        rpc.get_vlan_allocation.return_value = {
            'assignedVlans': '1-10,21-30',
            'availableVlans': '1-5,21,23,25,27,29',
            'allocatedVlans': '6-10,22,24,26,28,30'
        }

        assigned = list(itertools.chain(range(1, 11), range(21, 31)))

        available = [1, 2, 3, 4, 5, 21, 23, 25, 27, 29]
        allocated = list(set(assigned) - set(available))

        sync_service = VlanSyncService(rpc)
        sync_service.synchronize()

        self._ensure_in_db(assigned, allocated, available)

        # Call synchronize again which returns different data
        rpc.get_vlan_allocation.return_value = {
            'assignedVlans': '51-60,71-80',
            'availableVlans': '51-55,71,73,75,77,79',
            'allocatedVlans': '56-60,72,74,76,78,80'
        }

        assigned = list(itertools.chain(range(51, 61), range(71, 81)))

        available = [51, 52, 53, 54, 55, 71, 73, 75, 77, 79]
        allocated = list(set(assigned) - set(available))

        sync_service = VlanSyncService(rpc)
        sync_service.synchronize()

        self._ensure_in_db(assigned, allocated, available)



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from eventlet import greenthread
from eventlet import queue
import mock
import time

from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_utils import importutils

from neutron.tests.unit import testlib_api

from networking_arista.common import constants as a_const
from networking_arista.common import exceptions as arista_exc
from networking_arista.ml2 import arista_sync
from networking_arista.ml2.mechanism_arista import MechResource
from networking_arista.tests.unit import utils


class SyncServiceTest(testlib_api.SqlTestCase):
    """Test cases for the sync service."""

    def setUp(self):
        super(SyncServiceTest, self).setUp()
        utils.setup_arista_wrapper_config(cfg)
        plugin_klass = importutils.import_class(
            "neutron.db.db_base_plugin_v2.NeutronDbPluginV2")
        directory.add_plugin(plugin_constants.CORE, plugin_klass())
        utils.setup_scenario()
        self.mech_queue = queue.LightQueue()
        self.sync_service = arista_sync.AristaSyncWorker(self.mech_queue)
        self.sync_service._rpc = utils.MockCvx('region')

    def tearDown(self):
        if self.sync_service._running:
            self.sync_service.stop()
            self.sync_service.wait()
        super(SyncServiceTest, self).tearDown()

    def test_start(self):
        self.sync_service.start()
        self.assertTrue(self.sync_service._running)
        self.assertIsNotNone(self.sync_service._thread)
        self.assertIsNotNone(self.sync_service.sync_order)
        self.assertIsNotNone(self.sync_service.done)

    def test_start_twice(self):
        self.sync_service.start()
        current_thread = self.sync_service._thread
        self.sync_service.start()
        self.assertEqual(self.sync_service._thread, current_thread)

    def test_stop_graceful(self):
        self.test_start()
        running_thread = self.sync_service._thread
        self.sync_service.stop()
        self.sync_service.wait()
        self.assertFalse(self.sync_service._running)
        self.assertTrue(running_thread.dead)
        self.assertIsNone(self.sync_service._thread)

    def test_stop_ungraceful(self):
        self.test_start()
        running_thread = self.sync_service._thread
        self.sync_service.stop(graceful=False)
        self.assertTrue(running_thread.dead)
        self.assertIsNone(self.sync_service._thread)

    def test_reset(self):
        self.test_start()
        old_thread = self.sync_service._thread
        self.sync_service.reset()
        self.assertNotEqual(self.sync_service._thread, old_thread)

    def test_resource_class_full_coverage(self):
        self.sync_service.initialize()
        for i in range(len(self.sync_service.sync_order)):
            self.sync_service.sync_order[i] = mock.MagicMock()
        self.sync_service.synchronize_resources()
        for resource_type in self.sync_service.sync_order:
            resource_type.delete_cvx_resources.assert_called_once()
            resource_type.create_cvx_resources.assert_called_once()

    def test_full_sync_cvx_populated(self):
        self.sync_service.initialize()
        self.sync_service.synchronize_resources()
        for endpoint, data in self.sync_service._rpc.endpoint_data.items():
            self.assertNotEqual(data, {})

    def test_process_mech_update(self):
        self.sync_service.initialize()
        for resource_type in a_const.ALL_RESOURCE_TYPES:
            res_cls = mock.MagicMock()
            with mock.patch.object(self.sync_service,
                                   'get_resource_class') as get:
                get.return_value = res_cls
                res = MechResource('id', resource_type, a_const.CREATE)
                self.sync_service.update_neutron_resource(res)
                get.assert_called_once_with(resource_type)
                res_cls.update_neutron_resource.assert_called_once_with(
                    'id', a_const.CREATE)
                get.reset_mock()
                get.return_value = res_cls
                res = MechResource('id', resource_type, a_const.DELETE)
                self.sync_service.update_neutron_resource(res)
                get.assert_called_once_with(resource_type)
                res_cls.update_neutron_resource.assert_called_once_with(
                    'id', a_const.DELETE)

    def test_force_full_sync(self):
        self.sync_service.initialize()
        for i in range(len(self.sync_service.sync_order)):
            self.sync_service.sync_order[i] = mock.MagicMock()
        self.sync_service.force_full_sync()
        for resource_type in self.sync_service.sync_order:
            resource_type.clear_all_data.assert_called_once()

    def test_sync_timeout(self):
        self.sync_service.initialize()
        with mock.patch.object(
                self.sync_service, 'check_if_out_of_sync') as oos:
            self.sync_service.wait_for_sync_required()
            oos.assert_called_once()

    def test_full_sync_required(self):
        self.sync_service.initialize()
        self.sync_service._cvx_uuid = 'old-id'
        self.sync_service._rpc = mock.MagicMock()
        self.sync_service._rpc.get_cvx_uuid.return_value = 'new-id'
        with mock.patch.object(self.sync_service, 'force_full_sync') as ffs:
            self.assertTrue(self.sync_service.check_if_out_of_sync())
            ffs.assert_called_once()
            self.assertEqual(self.sync_service._synchronizing_uuid, 'new-id')
            self.assertNotEqual(self.sync_service._last_sync_time, 0)

    def test_mech_queue_timeout(self):
        self.sync_service.initialize()
        self.assertFalse(self.sync_service.wait_for_mech_driver_update(1))

    def test_mech_queue_updated(self):
        self.sync_service.initialize()
        resource = MechResource('tid', a_const.TENANT_RESOURCE, a_const.CREATE)
        self.mech_queue.put(resource)
        # Must yield to allow resource to be available on the queue
        greenthread.sleep(0)
        self.assertTrue(self.sync_service.wait_for_mech_driver_update(1))
        self.assertEqual(self.sync_service._resources_to_update, [resource])

    def test_sync_start_fail(self):
        self.sync_service.initialize()
        self.sync_service._rpc = mock.MagicMock()
        self.sync_service._rpc.sync_start.return_value = False
        self.assertEqual(self.sync_service._last_sync_time, 0)
        for i in range(len(self.sync_service.sync_order)):
            self.sync_service.sync_order[i] = mock.MagicMock()
        self.sync_service.synchronize_resources()
        for resource_type in self.sync_service.sync_order:
            resource_type.delete_cvx_resources.assert_not_called()
            resource_type.create_cvx_resources.assert_not_called()
        self.sync_service._rpc.sync_end.assert_not_called()

    def test_sync_lock_release_on_failure(self):
        self.sync_service.initialize()
        self.sync_service._rpc = mock.MagicMock()
        self.sync_service.update_neutron_resource = mock.Mock(
            side_effect=arista_exc.AristaRpcError(msg='fail'))
        resource = MechResource('tid', a_const.TENANT_RESOURCE, a_const.CREATE)
        self.sync_service.start()
        self.sync_service._last_sync_time = time.time()
        self.sync_service._cvx_uuid = 'cvx-id'
        self.mech_queue.put(resource)
        greenthread.sleep(0)
        self.sync_service.stop()
        self.sync_service.wait()
        self.sync_service._rpc.sync_end.assert_called_once()



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from neutron_lib import constants as n_const
from oslo_config import cfg

from networking_arista.tests.unit.ml2 import ml2_test_base

from neutron_lib.api.definitions import portbindings


class BasicMechDriverTestCase(ml2_test_base.MechTestBase):

    def test_create_network(self):
        # Test create regular network
        tenant_id = 'tid'
        reg_net_dict = {'network': {'name': 'net1',
                                    'tenant_id': tenant_id,
                                    'admin_state_up': True,
                                    'shared': False,
                                    'provider:physical_network': self.physnet,
                                    'provider:network_type': 'vlan'}}
        reg_network, reg_n_ctx = self.create_network(reg_net_dict)
        self.assertTenantCreated(tenant_id)
        self.assertNetworkCreated(reg_network['id'])
        for segment in reg_n_ctx.network_segments:
            self.assertSegmentCreated(segment['id'])

        # Test create shared network
        shrd_net_dict = {'network': {'name': 'shared_net',
                                     'tenant_id': tenant_id,
                                     'admin_state_up': True,
                                     'shared': True,
                                     'provider:physical_network': self.physnet,
                                     'provider:network_type': 'vlan'}}
        shared_network, shared_n_ctx = self.create_network(shrd_net_dict)
        self.assertTenantCreated(tenant_id)
        self.assertNetworkCreated(shared_network['id'])
        for segment in shared_n_ctx.network_segments:
            self.assertSegmentCreated(segment['id'])

        # Test delete regular network
        self.delete_network(reg_network['id'])
        self.assertTenantCreated(tenant_id)
        self.assertNetworkDeleted(reg_network['id'])
        for segment in reg_n_ctx.network_segments:
            self.assertSegmentDeleted(segment['id'])

        # Test delete shared network
        self.delete_network(shared_network['id'])
        self.assertTenantDeleted(tenant_id)
        self.assertNetworkDeleted(shared_network['id'])
        for segment in shared_n_ctx.network_segments:
            self.assertSegmentDeleted(segment['id'])

    def test_basic_dhcp_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)

        # Create DHCP port
        device_id = 'dhcp-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': n_const.DEVICE_OWNER_DHCP,
                     'binding:host_id': port_host}
        port, _ = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        self.assertDhcpCreated(device_id)
        self.assertDhcpPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host))

        # Delete DHCP port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertDhcpDeleted(device_id)
        self.assertDhcpPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))

    def test_basic_dvr_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)

        # Create DVR port
        device_id = 'router-1'
        port_tenant = 'port-ten'
        port_host_1 = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': n_const.DEVICE_OWNER_DVR_INTERFACE}
        port, _ = self.create_port(port_dict)
        self.bind_dvr_to_host(port, port_host_1)
        self.assertTenantCreated(port_tenant)
        self.assertRouterCreated(device_id)
        self.assertRouterPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host_1))

        # Bring up a second DVR host
        port_host_2 = self.host2
        port, port_ctx = self.bind_dvr_to_host(port, port_host_2)
        self.assertPortBindingCreated((port['id'], port_host_2))

        # Delete the port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertRouterDeleted(device_id)
        self.assertRouterPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host_1))
        self.assertPortBindingDeleted((port['id'], port_host_2))

    def test_dvr_port_host_bind_unbind(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)

        # Create DVR port
        device_id = 'router-1'
        port_tenant = 'port-ten'
        port_host_1 = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': n_const.DEVICE_OWNER_DVR_INTERFACE}
        port, _ = self.create_port(port_dict)
        port, port_ctx = self.bind_dvr_to_host(port, port_host_1)
        self.assertTenantCreated(port_tenant)
        self.assertRouterCreated(device_id)
        self.assertRouterPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host_1))

        # Bring up a second DVR host
        port_host_2 = self.host2
        port, port_ctx = self.bind_dvr_to_host(port, port_host_2)
        self.assertPortBindingCreated((port['id'], port_host_2))

        # Removed the second host
        self.unbind_dvr_from_host(port, port_host_2)
        self.assertPortBindingDeleted((port['id'], port_host_2))
        self.assertTenantCreated(port_tenant)
        self.assertRouterCreated(device_id)
        self.assertRouterPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host_1))

        # Delete the port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertRouterDeleted(device_id)
        self.assertRouterPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host_1))

    def test_basic_vm_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)

        # Create VM port
        device_id = 'vm-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': 'compute:',
                     'binding:host_id': port_host}
        port, _ = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        self.assertVmCreated(device_id)
        self.assertVmPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host))

        # Delete VM port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))

    def test_basic_baremetal_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)

        # Create baremetal port
        device_id = 'baremetal-1'
        port_tenant = 'port-ten'
        port_host = 'bm-host'
        switch_id = '00:11:22:33:44:55'
        switch_port = 'Ethernet1/1'
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': n_const.DEVICE_OWNER_BAREMETAL_PREFIX,
                     'binding:host_id': port_host,
                     'binding:profile': {'local_link_information': [
                         {'switch_id': switch_id,
                          'port_id': switch_port}]},
                     'binding:vnic_type': 'baremetal'}
        port, _ = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        self.assertBaremetalCreated(device_id)
        self.assertBaremetalPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], (switch_id, switch_port)))

        # Delete baremetal port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertBaremetalDeleted(device_id)
        self.assertBaremetalPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], (switch_id, switch_port)))

    def test_basic_baremetal_mlag(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)

        # Create baremetal port
        device_id = 'baremetal-1'
        port_tenant = 'port-ten'
        port_host = 'bm-host'
        switch_1_id = '00:11:22:33:44:55'
        switch_1_port = 'Ethernet1/1'
        switch_2_id = '55:44:33:22:11:00'
        switch_2_port = 'Ethernet2'
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': n_const.DEVICE_OWNER_BAREMETAL_PREFIX,
                     'binding:host_id': port_host,
                     'binding:profile': {'local_link_information': [
                         {'switch_id': switch_1_id,
                          'port_id': switch_1_port},
                         {'switch_id': switch_2_id,
                          'port_id': switch_2_port}]},
                     'binding:vnic_type': 'baremetal'}
        port, _ = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        self.assertBaremetalCreated(device_id)
        self.assertBaremetalPortCreated(port['id'])
        self.assertPortBindingCreated(
            (port['id'], (switch_1_id, switch_1_port)))
        self.assertPortBindingCreated(
            (port['id'], (switch_2_id, switch_2_port)))

        # Delete baremetal port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertBaremetalDeleted(device_id)
        self.assertBaremetalPortDeleted(port['id'])
        self.assertPortBindingDeleted(
            (port['id'], (switch_1_id, switch_2_port)))
        self.assertPortBindingDeleted(
            (port['id'], (switch_2_id, switch_2_port)))

    def test_host_migration(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)

        # Create VM port
        device_id = 'vm-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': 'compute:',
                     'binding:host_id': port_host}
        port, _ = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        self.assertVmCreated(device_id)
        self.assertVmPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host))

        # Migrate the port
        new_port_host = self.host2
        port, _ = self.migrate_port(port['id'], new_port_host)
        self.assertTenantCreated(port_tenant)
        self.assertVmCreated(device_id)
        self.assertVmPortCreated(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))
        self.assertPortBindingCreated((port['id'], new_port_host))

        # Delete VM port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], new_port_host))

    def test_dhcp_migration(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': 'physnet1',
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)

        # Create DHCP port
        device_id = 'dhcp-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': n_const.DEVICE_OWNER_DHCP,
                     'binding:host_id': port_host,
                     'binding:vnic_type': 'normal'}
        port, _ = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        self.assertDhcpCreated(device_id)
        self.assertDhcpPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host))

        # Migrate the DHCP port to a new dhcp instance
        new_device_id = 'dhcp-2'
        self.migrate_dhcp_device(port['id'], new_device_id)
        self.assertTenantCreated(port_tenant)
        self.assertDhcpCreated(new_device_id)
        self.assertDhcpDeleted(device_id)
        self.assertDhcpPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host))

        # Delete DHCP port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertDhcpDeleted(device_id)
        self.assertDhcpPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))

    def test_vm_trunk_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net1',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network1, _ = self.create_network(net_dict)
        net_dict = {'network': {'name': 'net2',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network2, _ = self.create_network(net_dict)

        # Create trunk port with subport
        device_id = 'vm-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        trunkport_dict = {'name': 'port1',
                          'tenant_id': port_tenant,
                          'network_id': network1['id'],
                          'admin_state_up': True,
                          'fixed_ips': [],
                          'device_id': '',
                          'device_owner': ''}
        trunkport, _ = self.create_port(trunkport_dict)
        subport_dict = {'name': 'port2',
                        'tenant_id': port_tenant,
                        'network_id': network2['id'],
                        'admin_state_up': True,
                        'fixed_ips': [],
                        'device_id': '',
                        'device_owner': ''}
        subport, _ = self.create_port(subport_dict)
        trunk_dict = {'trunk': {'port_id': trunkport['id'],
                                'tenant_id': port_tenant,
                                'sub_ports': [{'port_id': subport['id'],
                                               'segmentation_type': 'vlan',
                                               'segmentation_id': 123}]}}
        trunk = self.trunk_plugin.create_trunk(self.context, trunk_dict)
        self.bind_trunk_to_host(trunkport, device_id, port_host)
        self.assertTenantCreated(port_tenant)
        self.assertVmCreated(device_id)
        self.assertVmPortCreated(trunkport['id'])
        self.assertPortBindingCreated((trunkport['id'], port_host))
        self.assertVmPortCreated(subport['id'])
        self.assertPortBindingCreated((subport['id'], port_host))

        # Delete the trunk and subport
        self.unbind_port_from_host(trunkport['id'])
        self.trunk_plugin.delete_trunk(self.context, trunk['id'])
        self.delete_port(trunkport['id'])
        self.delete_port(subport['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(trunkport['id'])
        self.assertPortBindingDeleted((trunkport['id'], port_host))
        self.assertVmPortDeleted(subport['id'])
        self.assertPortBindingDeleted((subport['id'], port_host))

    def test_trunk_add_remove_subport(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net1',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network1, _ = self.create_network(net_dict)
        net_dict = {'network': {'name': 'net2',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network2, _ = self.create_network(net_dict)
        net_dict = {'network': {'name': 'net3',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network3, _ = self.create_network(net_dict)

        # Create trunk port with subport, add subport after initial binding
        device_id = 'vm-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        trunkport_dict = {'name': 'port1',
                          'tenant_id': port_tenant,
                          'network_id': network1['id'],
                          'admin_state_up': True,
                          'fixed_ips': [],
                          'device_id': '',
                          'device_owner': ''}
        trunkport, _ = self.create_port(trunkport_dict)
        subport_dict = {'name': 'port2',
                        'tenant_id': port_tenant,
                        'network_id': network2['id'],
                        'admin_state_up': True,
                        'fixed_ips': [],
                        'device_id': '',
                        'device_owner': ''}
        subport, _ = self.create_port(subport_dict)
        trunk_dict = {'trunk': {'port_id': trunkport['id'],
                                'tenant_id': port_tenant,
                                'sub_ports': [{'port_id': subport['id'],
                                               'segmentation_type': 'vlan',
                                               'segmentation_id': 123}]}}
        subport_dict2 = {'name': 'port3',
                         'tenant_id': port_tenant,
                         'network_id': network3['id'],
                         'admin_state_up': True,
                         'fixed_ips': [],
                         'device_id': '',
                         'device_owner': ''}
        trunk = self.trunk_plugin.create_trunk(self.context, trunk_dict)
        self.bind_trunk_to_host(trunkport, device_id, port_host)
        subport2, _ = self.create_port(subport_dict2)
        self.trunk_plugin.add_subports(self.context, trunk['id'],
                                       {'sub_ports':
                                        [{'port_id': subport2['id'],
                                          'segmentation_type': 'vlan',
                                          'segmentation_id': 111}]})
        self.bind_subport_to_trunk(subport2, trunk)
        self.assertTenantCreated(port_tenant)
        self.assertVmCreated(device_id)
        self.assertVmPortCreated(trunkport['id'])
        self.assertPortBindingCreated((trunkport['id'], port_host))
        self.assertVmPortCreated(subport['id'])
        self.assertPortBindingCreated((subport['id'], port_host))
        self.assertVmPortCreated(subport2['id'])
        self.assertPortBindingCreated((subport2['id'], port_host))

        # Remove the trunk subport
        self.trunk_plugin.remove_subports(self.context, trunk['id'],
                                          {'sub_ports':
                                           [{'port_id': subport2['id']}]})
        self.unbind_port_from_host(subport2['id'])
        self.assertPortBindingDeleted((subport2['id'], port_host))

        # Delete the trunk and remaining subport
        self.unbind_port_from_host(trunkport['id'])
        self.trunk_plugin.delete_trunk(self.context, trunk['id'])
        self.delete_port(trunkport['id'])
        self.delete_port(subport['id'])
        self.delete_port(subport2['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(trunkport['id'])
        self.assertPortBindingDeleted((trunkport['id'], port_host))
        self.assertVmPortDeleted(subport['id'])
        self.assertPortBindingDeleted((subport['id'], port_host))

    def test_baremetal_trunk_basic(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net1',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network1, _ = self.create_network(net_dict)
        net_dict = {'network': {'name': 'net2',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network2, _ = self.create_network(net_dict)

        # Create baremetal port
        device_id = 'baremetal-1'
        port_tenant = 'port-ten'
        port_host = 'bm-host'
        switch_id = '00:11:22:33:44:55'
        switch_port = 'Ethernet1/1'
        trunkport_dict = {'name': 'port1',
                          'tenant_id': port_tenant,
                          'network_id': network1['id'],
                          'admin_state_up': True,
                          'fixed_ips': [],
                          'device_id': '',
                          'device_owner': ''}
        trunkport, _ = self.create_port(trunkport_dict)
        subport_dict = {'name': 'port2',
                        'tenant_id': port_tenant,
                        'network_id': network2['id'],
                        'admin_state_up': True,
                        'fixed_ips': [],
                        'device_id': '',
                        'device_owner': ''}
        subport, _ = self.create_port(subport_dict)
        trunk_dict = {'trunk': {'port_id': trunkport['id'],
                                'tenant_id': port_tenant,
                                'sub_ports': [{'port_id': subport['id'],
                                               'segmentation_type': 'inherit',
                                               'segmentation_id': 'inherit'}]}}
        self.trunk_plugin.create_trunk(self.context, trunk_dict)
        self.bind_trunk_to_baremetal(trunkport['id'], device_id, port_host,
                                     switch_id, switch_port)
        self.assertTenantCreated(port_tenant)
        self.assertBaremetalCreated(device_id)
        self.assertBaremetalPortCreated(trunkport['id'])
        self.assertPortBindingCreated(
            (trunkport['id'], (switch_id, switch_port)))
        self.assertBaremetalPortCreated(subport['id'])
        self.assertPortBindingCreated(
            (subport['id'], (switch_id, switch_port)))

        # Simulate baremetal shutdown
        self.unbind_trunk_from_baremetal(trunkport['id'])
        self.assertBaremetalDeleted(device_id)
        self.assertPortBindingDeleted(
            (trunkport['id'], (switch_id, switch_port)))
        self.assertPortBindingDeleted(
            (subport['id'], (switch_id, switch_port)))

    def test_baremetal_trunk_bind_unbind(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net1',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network1, _ = self.create_network(net_dict)
        net_dict = {'network': {'name': 'net2',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network2, _ = self.create_network(net_dict)
        net_dict = {'network': {'name': 'net3',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network3, _ = self.create_network(net_dict)

        # Create baremetal port
        device_id = 'baremetal-1'
        port_tenant = 'port-ten'
        port_host = 'bm-host'
        switch_id = '00:11:22:33:44:55'
        switch_port = 'Ethernet1/1'
        trunkport_dict = {'name': 'port1',
                          'tenant_id': port_tenant,
                          'network_id': network1['id'],
                          'admin_state_up': True,
                          'fixed_ips': [],
                          'device_id': '',
                          'device_owner': ''}
        trunkport, _ = self.create_port(trunkport_dict)
        subport_dict = {'name': 'port2',
                        'tenant_id': port_tenant,
                        'network_id': network2['id'],
                        'admin_state_up': True,
                        'fixed_ips': [],
                        'device_id': '',
                        'device_owner': ''}
        subport, _ = self.create_port(subport_dict)
        trunk_dict = {'trunk': {'port_id': trunkport['id'],
                                'tenant_id': port_tenant,
                                'sub_ports': [{'port_id': subport['id'],
                                               'segmentation_type': 'inherit',
                                               'segmentation_id': 'inherit'}]}}
        trunk = self.trunk_plugin.create_trunk(self.context, trunk_dict)
        self.bind_trunk_to_baremetal(trunkport['id'], device_id, port_host,
                                     switch_id, switch_port)
        self.assertTenantCreated(port_tenant)
        self.assertBaremetalCreated(device_id)
        self.assertBaremetalPortCreated(trunkport['id'])
        self.assertPortBindingCreated(
            (trunkport['id'], (switch_id, switch_port)))
        self.assertBaremetalPortCreated(subport['id'])
        self.assertPortBindingCreated(
            (subport['id'], (switch_id, switch_port)))

        subport_dict2 = {'name': 'port3',
                         'tenant_id': port_tenant,
                         'network_id': network3['id'],
                         'admin_state_up': True,
                         'fixed_ips': [],
                         'device_id': '',
                         'device_owner': ''}
        subport2, _ = self.create_port(subport_dict2)
        self.trunk_plugin.add_subports(self.context, trunk['id'],
                                       {'sub_ports':
                                        [{'port_id': subport2['id'],
                                          'segmentation_type': 'inherit',
                                          'segmentation_id': 'inherit'}]})
        self.assertBaremetalPortCreated(subport2['id'])
        self.assertPortBindingCreated(
            (subport2['id'], (switch_id, switch_port)))

        self.trunk_plugin.remove_subports(self.context, trunk['id'],
                                          {'sub_ports':
                                           [{'port_id': subport2['id']}]})
        self.assertPortBindingDeleted(
            (subport2['id'], (switch_id, switch_port)))

        # Simulate baremetal shutdown
        self.unbind_trunk_from_baremetal(trunkport['id'])
        self.assertBaremetalDeleted(device_id)
        self.assertPortBindingDeleted(
            (trunkport['id'], (switch_id, switch_port)))
        self.assertPortBindingDeleted(
            (subport['id'], (switch_id, switch_port)))

    def test_baremetal_trunk_pre_bound(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net1',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network1, _ = self.create_network(net_dict)
        net_dict = {'network': {'name': 'net2',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network2, _ = self.create_network(net_dict)

        # Create baremetal port
        device_id = 'baremetal-1'
        port_tenant = 'port-ten'
        port_host = 'bm-host'
        switch_id = '00:11:22:33:44:55'
        switch_port = 'Ethernet1/1'
        trunkport_dict = {'name': 'port1',
                          'tenant_id': port_tenant,
                          'network_id': network1['id'],
                          'admin_state_up': True,
                          'fixed_ips': [],
                          'device_id': '',
                          'device_owner': ''}
        trunkport, _ = self.create_port(trunkport_dict)
        subport_dict = {'name': 'port2',
                        'tenant_id': port_tenant,
                        'network_id': network2['id'],
                        'admin_state_up': True,
                        'fixed_ips': [],
                        'device_id': '',
                        'device_owner': ''}
        subport, _ = self.create_port(subport_dict)
        trunk_dict = {'trunk': {'port_id': trunkport['id'],
                                'tenant_id': port_tenant,
                                'sub_ports': [{'port_id': subport['id'],
                                               'segmentation_type': 'inherit',
                                               'segmentation_id': 'inherit'}]}}
        self.bind_trunk_to_baremetal(trunkport['id'], device_id, port_host,
                                     switch_id, switch_port)
        self.trunk_plugin.create_trunk(self.context, trunk_dict)
        self.assertTenantCreated(port_tenant)
        self.assertBaremetalCreated(device_id)
        self.assertBaremetalPortCreated(trunkport['id'])
        self.assertPortBindingCreated(
            (trunkport['id'], (switch_id, switch_port)))
        self.assertBaremetalPortCreated(subport['id'])
        self.assertPortBindingCreated(
            (subport['id'], (switch_id, switch_port)))

        # Simulate baremetal shutdown
        self.unbind_trunk_from_baremetal(trunkport['id'])
        self.assertBaremetalDeleted(device_id)
        self.assertPortBindingDeleted(
            (trunkport['id'], (switch_id, switch_port)))
        self.assertPortBindingDeleted(
            (subport['id'], (switch_id, switch_port)))

    def test_error_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': self.physnet,
                                'provider:network_type': 'vlan'}}
        network, _ = self.create_network(net_dict)

        # Create port
        device_id = 'vm-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': 'compute:',
                     'binding:host_id': port_host}
        port, _ = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        self.assertVmCreated(device_id)
        self.assertVmPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host))

        # Set port to ERROR state
        self.set_port_to_error_state(port)
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))


class BasicHpbMechDriverTestCase(ml2_test_base.MechTestBase):

    def setUp(self):
        cfg.CONF.set_override('manage_fabric', True, "ml2_arista")
        super(BasicHpbMechDriverTestCase, self).setUp()

    def test_basic_hpb_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': None,
                                'provider:network_type': 'vxlan'}}
        network, _ = self.create_network(net_dict)
        self.assertNetworkCreated(network['id'])

        # Create HPB port
        device_id = 'vm-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': 'compute:',
                     'binding:host_id': port_host}
        with mock.patch.object(self.drv.eapi,
                               'get_host_physnet',
                               return_value=self.physnet):
            port, port_ctx = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        self.assertVmCreated(device_id)
        self.assertVmPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host))

        # Check that the dynamic segment was created
        network_segments = [level['bound_segment']
                            for level in port_ctx.binding_levels]
        self.assertTrue(len(network_segments) == 2)
        for segment in network_segments:
            self.assertSegmentCreated(segment['id'])

        # Delete HPB port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))

    def test_hpb_dvr_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': None,
                                'provider:network_type': 'vxlan'}}
        network, _ = self.create_network(net_dict)

        # Create DVR port
        device_id = 'router-1'
        port_tenant = 'port-ten'
        port_host_1 = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': n_const.DEVICE_OWNER_DVR_INTERFACE}
        port, _ = self.create_port(port_dict)
        with mock.patch.object(self.drv.eapi,
                               'get_host_physnet',
                               return_value=self.physnet):
            port, port_ctx = self.bind_dvr_to_host(port, port_host_1)
        self.assertTenantCreated(port_tenant)
        self.assertRouterCreated(device_id)
        self.assertRouterPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host_1))
        network_segments = [level['bound_segment']
                            for level in port_ctx.binding_levels]
        self.assertTrue(len(network_segments) == 2)
        for segment in network_segments:
            self.assertSegmentCreated(segment['id'])

        # Bring up a second DVR host
        port_host_2 = self.host3
        with mock.patch.object(self.drv.eapi,
                               'get_host_physnet',
                               return_value=self.physnet2):
            port, port_ctx = self.bind_dvr_to_host(port, port_host_2)
        self.assertPortBindingCreated((port['id'], port_host_2))
        network_segments = [level['bound_segment']
                            for level in port_ctx.binding_levels]
        self.assertTrue(len(network_segments) == 2)
        for segment in network_segments:
            self.assertSegmentCreated(segment['id'])

        # Delete the port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertRouterDeleted(device_id)
        self.assertRouterPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host_1))
        self.assertPortBindingDeleted((port['id'], port_host_2))


class UnmanagedFabricUnmanagedPhysnetHpbTestCase(ml2_test_base.MechTestBase):

    _mechanism_drivers = ['arista_test_fabric', 'arista', 'openvswitch']

    def setUp(self):
        cfg.CONF.set_override('manage_fabric', False, "ml2_arista")
        cfg.CONF.set_override('managed_physnets', ['other_physnet'],
                              "ml2_arista")
        super(UnmanagedFabricUnmanagedPhysnetHpbTestCase, self).setUp()

    def test_unmanaged_fabric_unmanaged_hpb_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': None,
                                'provider:network_type': 'vxlan'}}
        network, _ = self.create_network(net_dict)
        self.assertNetworkCreated(network['id'])

        # Create HPB port
        device_id = 'vm-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': 'compute:',
                     'binding:host_id': port_host}
        port, port_ctx = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        # Check that the dynamic segment was created
        network_segments = [level['bound_segment']
                            for level in port_ctx.binding_levels]
        self.assertTrue(len(network_segments) == 2)
        for segment in network_segments:
            self.assertSegmentCreated(segment['id'])

        # The VM/Port should not have been created
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))

        # Delete HPB port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))


class ManagedFabricUnmanagedPhysnetHpbTestCase(ml2_test_base.MechTestBase):

    def setUp(self):
        cfg.CONF.set_override('manage_fabric', True, "ml2_arista")
        cfg.CONF.set_override('managed_physnets', ['other_physnet'],
                              "ml2_arista")
        super(ManagedFabricUnmanagedPhysnetHpbTestCase, self).setUp()

    def test_managed_fabric_unmanaged_hpb_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': None,
                                'provider:network_type': 'vxlan'}}
        network, _ = self.create_network(net_dict)
        self.assertNetworkCreated(network['id'])

        # Create HPB port
        device_id = 'vm-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': 'compute:',
                     'binding:host_id': port_host}
        with mock.patch.object(self.drv.eapi,
                               'get_host_physnet',
                               return_value=self.physnet):
            port, port_ctx = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        # Check that the dynamic segment was created
        network_segments = [level['bound_segment']
                            for level in port_ctx.binding_levels]
        self.assertTrue(len(network_segments) == 2)
        for segment in network_segments:
            self.assertSegmentCreated(segment['id'])

        # The VM/Port should not have been created
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))

        # Delete HPB port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))


class UnmanagedFabricManagedPhysnetHpbTestCase(ml2_test_base.MechTestBase):

    _mechanism_drivers = ['arista_test_fabric', 'arista', 'openvswitch']

    def setUp(self):
        self.physnet = 'physnet1'
        cfg.CONF.set_override('manage_fabric', False, "ml2_arista")
        cfg.CONF.set_override('managed_physnets', [self.physnet],
                              "ml2_arista")
        super(UnmanagedFabricManagedPhysnetHpbTestCase, self).setUp()

    def test_unmanaged_fabric_managed_hpb_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': None,
                                'provider:network_type': 'vxlan'}}
        network, _ = self.create_network(net_dict)
        self.assertNetworkCreated(network['id'])

        # Create HPB port
        device_id = 'vm-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': 'compute:',
                     'binding:host_id': port_host}
        port, port_ctx = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        self.assertVmCreated(device_id)
        self.assertVmPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host))

        # Check that the dynamic segment was created
        network_segments = [level['bound_segment']
                            for level in port_ctx.binding_levels]
        self.assertTrue(len(network_segments) == 2)
        for segment in network_segments:
            self.assertSegmentCreated(segment['id'])

        # Delete HPB port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))


class ManagedFabricManagedFabricHpbTestCase(ml2_test_base.MechTestBase):

    def setUp(self):
        self.physnet = 'physnet1'
        cfg.CONF.set_override('manage_fabric', True, "ml2_arista")
        cfg.CONF.set_override('managed_physnets', [self.physnet],
                              "ml2_arista")
        super(ManagedFabricManagedFabricHpbTestCase, self).setUp()

    def test_managed_fabric_managed_hpb_port(self):
        network_tenant = 'net-ten'
        net_dict = {'network': {'name': 'net',
                                'tenant_id': network_tenant,
                                'admin_state_up': True,
                                'shared': False,
                                'provider:physical_network': None,
                                'provider:network_type': 'vxlan'}}
        network, _ = self.create_network(net_dict)
        self.assertNetworkCreated(network['id'])

        # Create HPB port
        device_id = 'vm-1'
        port_tenant = 'port-ten'
        port_host = self.host1
        port_dict = {'name': 'port1',
                     'tenant_id': port_tenant,
                     'network_id': network['id'],
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': device_id,
                     'device_owner': 'compute:',
                     'binding:host_id': port_host}
        with mock.patch.object(self.drv.eapi,
                               'get_host_physnet',
                               return_value=self.physnet):
            port, port_ctx = self.create_port(port_dict)
        self.assertTenantCreated(port_tenant)
        self.assertVmCreated(device_id)
        self.assertVmPortCreated(port['id'])
        self.assertPortBindingCreated((port['id'], port_host))

        # Check that the dynamic segment was created
        network_segments = [level['bound_segment']
                            for level in port_ctx.binding_levels]
        self.assertTrue(len(network_segments) == 2)
        for segment in network_segments:
            self.assertSegmentCreated(segment['id'])

        # Delete HPB port
        self.delete_port(port['id'])
        self.assertTenantDeleted(port_tenant)
        self.assertVmDeleted(device_id)
        self.assertVmPortDeleted(port['id'])
        self.assertPortBindingDeleted((port['id'], port_host))


class BasicL3HARouterTests(object):

    def test_create_delete_router(self):
        ha_router_ports = []
        router = self.create_router(ha=True)
        for l3_agent in self.l3_agents:
            port = self.update_routers_states(router['id'], l3_agent)
            ha_router_ports.append(port)
        ha_network_id = self.get_ha_network(router)
        ha_segments = self.get_network_segments(ha_network_id)
        self.assertTenantCreated(router['project_id'])
        self.assertL3HANetworkCreated(router, ha_network_id)
        self.assertRouterCreated(router['id'])
        self.assertL3HARouterCreated(router)
        for port in ha_router_ports:
            self.assertL3HAPortCreated(router, port['id'])
            self.assertPortBindingCreated((port['id'],
                                           port[portbindings.HOST_ID]))
        self.assertSegmentsCreated(ha_segments)

        # Delete the router
        self.delete_router(router['id'])
        self.assertRouterPortsDeleted([p['id'] for p in ha_router_ports])
        self.assertRouterDeleted(router['id'])
        self.assertSegmentsDeleted(ha_segments)
        self.assertNetworkDeleted(ha_network_id)
        self.assertTenantDeleted(router['project_id'])
        for port in ha_router_ports:
            self.assertPortBindingDeleted((port['id'],
                                           port[portbindings.HOST_ID]))


class BasicRouterTests(object):
    def test_create_delete_router(self):
        router = self.create_router()
        net_list = []
        port_list = []
        segment_list = []
        for net in self.net_dict:
            network, net_ctx = self.create_network(net)
            net_list.append((network, net_ctx))
        self.assertTenantCreated(router['project_id'])
        for net, net_ctx in net_list:
            interface_info = {'subnet_id': net_ctx.current['subnets'][0]}
            intf = self.add_router_interface(router, interface_info)
            self.sync_routers(router['id'], self.l3_agent1['host'])
            port = self.get_legacy_router_port(intf['port_id'])
            self.assertNotEqual(len(port), 0)
            port_list.append(port)
        self.assertLegacyRouterCreated(router, self.l3_agent1['host'])
        for network, _ in net_list:
            self.assertNetworkCreated(network['id'])
            segment_list.extend(self.get_network_segments(network['id']))
        self.assertEqual(len(segment_list), self.total_segments)
        self.assertSegmentsCreated(segment_list)
        for port in port_list:
            self.assertRouterPortCreated(port['id'])
            self.assertPortBindingCreated((port['id'],
                                           port[portbindings.HOST_ID]))

        # Delete the router interfaces and router
        # Remove one of router's interface
        network, net_ctx = net_list[0]
        interface_info = {'subnet_id': net_ctx.current['subnets'][0]}
        intf = self.remove_router_interface(router, interface_info)
        self.assertRouterCreated(router['id'])
        self.assertRouterPortDeleted(intf['port_id'])
        self.assertPortBindingDeleted((intf['port_id'],
                                       port[portbindings.HOST_ID]))

        # Remove second router interface
        network, net_ctx = net_list[1]
        interface_info = {'subnet_id': net_ctx.current['subnets'][0]}
        intf = self.remove_router_interface(router, interface_info)
        self.assertRouterDeleted(router['id'])
        self.assertRouterPortDeleted(intf['port_id'])
        self.assertPortBindingDeleted((intf['port_id'],
                                       port[portbindings.HOST_ID]))

        for network, _ in net_list:
            self.delete_network(network['id'])
            self.assertNetworkDeleted(network['id'])
        self.assertSegmentsDeleted(segment_list)


class BasicL3HARouterTestCases(ml2_test_base.L3HARouterTestFramework,
                               BasicL3HARouterTests):

    def setUp(self):
        cfg.CONF.set_override('tenant_network_types', 'vlan', 'ml2')
        super(BasicL3HARouterTestCases, self).setUp()
        cfg.CONF.set_override('max_l3_agents_per_router', 2)
        self.l3_agent1 = self._register_l3_agent(host=self.host1)
        self.l3_agent2 = self._register_l3_agent(host=self.host2)
        self.l3_agents = [self.l3_agent1, self.l3_agent2]


class BasicHpbL3HARouterTestCases(ml2_test_base.L3HARouterTestFramework,
                                  BasicL3HARouterTests):
    def setUp(self):
        cfg.CONF.set_override('manage_fabric', True, 'ml2_arista')
        cfg.CONF.set_override('tenant_network_types', 'vxlan', 'ml2')
        super(BasicHpbL3HARouterTestCases, self).setUp()
        cfg.CONF.set_override('l3_ha', True)
        cfg.CONF.set_override('max_l3_agents_per_router', 3)
        self.l3_agent1 = self._register_l3_agent(host=self.host1)
        self.l3_agent2 = self._register_l3_agent(host=self.host2)
        self.l3_agent3 = self._register_l3_agent(host=self.host3)
        self.l3_agents = [self.l3_agent1, self.l3_agent2, self.l3_agent3]

        def get_host_physnet(context):
            if context.host == 'host3':
                return self.physnet2
            if context.host == 'host1':
                return self.physnet
            if context.host == 'host2':
                return self.physnet

        ghp = mock.patch.object(self.drv.eapi, 'get_host_physnet').start()
        ghp.side_effect = get_host_physnet


class BasicRouterTestCases(ml2_test_base.L3HARouterTestFramework,
                           BasicRouterTests):

    def setUp(self):
        cfg.CONF.set_override('tenant_network_types', 'vlan', 'ml2')
        super(BasicRouterTestCases, self).setUp()
        self.l3_agent1 = self._register_l3_agent(host=self.host1)
        self.net_dict = [
            {'network': {'name': 'net-%d' % r,
                         'tenant_id': self._tenant_id,
                         'admin_state_up': True,
                         'shared': False,
                         'provider:physical_network': self.physnet,
                         'provider:network_type': 'vlan'}}
            for r in range(1, 3)]
        self.total_segments = 2
        self.l3_agents = [self.l3_agent1]


class BasicHpbRouterTestCases(ml2_test_base.L3HARouterTestFramework,
                              BasicRouterTests):

    def setUp(self):
        cfg.CONF.set_override('manage_fabric', True, 'ml2_arista')
        cfg.CONF.set_override('tenant_network_types', 'vxlan', 'ml2')
        super(BasicHpbRouterTestCases, self).setUp()
        self.l3_agent1 = self._register_l3_agent(host=self.host1)

        def get_host_physnet(context):
            return self.physnet
        ghp = mock.patch.object(self.drv.eapi, 'get_host_physnet').start()
        ghp.side_effect = get_host_physnet
        self.net_dict = [{'network': {'name': 'hpb_net-%d' % r,
                                      'tenant_id': self._tenant_id,
                                      'admin_state_up': True,
                                      'shared': False}}
                         for r in range(1, 3)]
        self.total_segments = 4
        self.l3_agents = [self.l3_agent1]



# Copyright (c) 2018 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib import worker
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall

from networking_arista.ml2.security_groups import switch_helper

LOG = logging.getLogger(__name__)


class AristaSecurityGroupSyncWorker(
        worker.BaseWorker,
        switch_helper.AristaSecurityGroupSyncHelper):
    """Worker that handles synchronizing Security Group ACLs on Arista switches

    The worker periodically queries the neutron db and sends all security
    groups, security group rules and security group port bindings to to
    registered switches.
    """

    def __init__(self):
        super(AristaSecurityGroupSyncWorker, self).__init__()
        self.initialize_switch_endpoints()
        self._loop = None

    def start(self):
        super(AristaSecurityGroupSyncWorker, self).start()
        if self._loop is None:
            self._loop = loopingcall.FixedIntervalLoopingCall(
                self.synchronize
            )
        self._loop.start(interval=cfg.CONF.ml2_arista.sync_interval)

    def stop(self):
        if self._loop is not None:
            self._loop.stop()

    def wait(self):
        if self._loop is not None:
            self._loop.wait()
        self._loop = None

    def reset(self):
        self.stop()
        self.wait()
        self.start()

    def synchronize_switch(self, switch_ip, expected_acls, expected_bindings):
        """Update ACL config on a switch to match expected config

        This is done as follows:
        1. Get switch ACL config using show commands
        2. Update expected bindings based on switch LAGs
        3. Get commands to synchronize switch ACLs
        4. Get commands to synchronize switch ACL bindings
        5. Run sync commands on switch
        """

        # Get ACL rules and interface mappings from the switch
        switch_acls, switch_bindings = self._get_dynamic_acl_info(switch_ip)

        # Adjust expected bindings for switch LAG config
        expected_bindings = self.adjust_bindings_for_lag(switch_ip,
                                                         expected_bindings)
        # Get synchronization commands
        switch_cmds = list()
        switch_cmds.extend(
            self.get_sync_acl_cmds(switch_acls, expected_acls))
        switch_cmds.extend(
            self.get_sync_binding_cmds(switch_bindings, expected_bindings))

        # Update switch config
        self.run_openstack_sg_cmds(switch_cmds, self._switches.get(switch_ip))

    def synchronize(self):
        """Perform sync of the security groups between ML2 and EOS."""

        # Get expected ACLs and rules
        expected_acls = self.get_expected_acls()

        # Get expected interface to ACL mappings
        all_expected_bindings = self.get_expected_bindings()

        # Check that config is correct on every registered switch
        for switch_ip in self._switches.keys():
            expected_bindings = all_expected_bindings.get(switch_ip, [])
            try:
                self.synchronize_switch(switch_ip, expected_acls,
                                        expected_bindings)
            except Exception:
                LOG.exception("Failed to sync SGs for %(switch)s",
                              {'switch': switch_ip})
